

__all__ = ['set_formatter', 'set_info_options', 'set_convert_options']

import os
from optparse import OptionGroup, NO_DEFAULT
from optparse import TitledHelpFormatter

try:
    import wx
    have_wx = True
except ImportError:
    have_wx = False

class MyHelpFormatter(TitledHelpFormatter):

    def format_option(self, option):
        old_help = option.help
        default = option.default
        if isinstance (default, str) and ' ' in default:
            default = repr (default)
        if option.help is None:
            option.help = 'Specify a %s.' % (option.type) 
        if option.type=='choice':
            choices = []
            for choice in option.choices:
                if choice==option.default:
                    if ' ' in choice:
                        choice = repr(choice)
                    choice = '['+choice+']'
                else:
                    if ' ' in choice:
                        choice = repr(choice)
                choices.append (choice)
            option.help = '%s Choices: %s.'% (option.help, ', '.join(choices))
        else:
            if default != NO_DEFAULT:
                if option.action=='store_false':
                    option.help = '%s Default: %s.'% (option.help, not default)
                else:
                    option.help = '%s Default: %s.'% (option.help, default)

        result = TitledHelpFormatter.format_option (self, option)
        option.help = old_help
        return result

help_formatter = MyHelpFormatter()

def set_formatter(parser):
    """Set customized help formatter.
    """
    parser.formatter = help_formatter

def set_convert_options(parser):
    set_formatter(parser)
    if os.name == 'posix':
        try:
            import matplotlib
            matplotlib.use('GTkAgg')
            parser.run_methods = ['subcommand']
        except ImportError:
            pass
    parser.set_usage ('%prog [options] -i INPUTPATH [-o OUTPUTPATH]')
    parser.set_description('Convert INPUTPATH to OUTPUTPATH.')
    parser.add_option ('--input-path', '-i',
                       type = 'file' if have_wx else str, metavar='INPUTPATH',
                       help = 'Specify INPUTPATH.'
                       )
    parser.add_option ('--output-path', '-o',
                       type = 'file' if have_wx else str, metavar='OUTPUTPATH',
                       help = 'Specify OUTPUTPATH.'
                       )
    parser.add_option ('--compression',
                       type = 'choice', default='none',
                       choices = ['none', 'lzw'],
                       help = 'Specify compression.'
                       )
    parser.add_option ('--slice',
                       type = 'string',
                       help = 'Specify slice using form "<zstart>:<zend>,<ystart>:<yend>,<xstart>:<xend>"'
                       )

def set_info_options(parser):
    set_formatter(parser)
    if os.name == 'posix':
        try:
            import matplotlib
            matplotlib.use('GTkAgg')
            parser.run_methods = ['subcommand']
        except ImportError:
            pass
    parser.set_usage ('%prog [options] -i INPUTPATH')
    parser.set_description('Show INPUTPATHs information.')
    parser.add_option ('--input-path', '-i',
                       type = 'file' if have_wx else str, metavar='INPUTPATH',
                       help = 'Specify INPUTPATH.'
                       )
    parser.add_option ('--memory-usage',
                       action = 'store_true', default=False,
                       help = 'Show TIFF file memory usage.'
                       )
    parser.add_option ('--no-memory-usage', dest='memory_usage',
                       action = 'store_false',
                       help = 'See --memory-usage.'
                       )
    parser.add_option ('--ifd',
                       action = 'store_true', default=False,
                       help = 'Show all TIFF file image file directory. By default, only the first IFD is shown.'
                       )
    parser.add_option ('--no-ifd', dest='ifd',
                       action = 'store_false', help='See --ifd.')
 
    parser.add_option ('--human',
                       action = 'store_true', default=False,
                       help = 'Show human readable values'
                       )
    parser.add_option ('--no-human', dest='human',
                       action = 'store_false', help='See --human.')



""" Implements an array of TIFF sample images.
"""
# Author: Pearu Peterson
# Created: Nov 2010


import sys
import numpy

__all__ = ['TiffArray']

class TiffArray:
    """ Holds a sequence of homogeneous TiffPlane's.

    TiffPlane's are homogeneous if they contain sample images of same
    shape and type. Otherwise TiffPlane's may contain images from
    different TIFF files with different pixel content.
    """
    def __init__ (self, planes):
        self.planes = []
        self.shape = ()
        self.dtype = None
        list(map(self.append, planes))

    def __len__(self):
        return self.shape[0]

    def __iter__(self):
        for plane in self.planes:
            yield plane

    def __getitem__ (self, index):
        try:
            if isinstance(index, int):
                if self.sample_index is None:
                    print(self.shape)
                return self.planes[index][()]
            elif isinstance (index, slice):
                indices = list(range(*index.indices(self.shape[0])))
                r = numpy.empty((len(indices),)+self.shape[1:], dtype=self.dtype)
                for i,j in enumerate(indices):
                    r[i] = self.planes[j][()]
                return r
            elif isinstance(index, tuple):
                if len (index)==0:
                    return self[:]
                if len (index)==1:
                    return self[index[0]]
                index0 = index[0]
                if isinstance(index0, int):
                    return self.planes[index0][index[1:]]
                elif isinstance (index0, slice):
                    indices = list(range(*index0.indices(self.shape[0])))
                    for i,j in enumerate(indices):
                        s = self.planes[j][index[1:]]
                        if i==0:
                            r = numpy.empty((len(indices),)+s.shape, dtype=self.dtype)
                        r[i] = s
                    return r
        except IOError as msg:
            sys.stderr.write('%s.__getitem__:\n%s\n' % (self.__class__.__name__, msg))
            sys.stderr.flush ()
            return None
        raise NotImplementedError (repr(index))

    def append(self, plane):
        """ Append tiff plane to tiff array.
        """
        if self.planes:
            if not self.planes[0].check_same_shape_and_type (plane):
                raise TypeError('planes are not homogeneous (same shape and sample type), expected %s but got %s' % ((self.planes[0].shape, self.dtype), (plane.shape, plane.dtype)))
            self.shape = (self.shape[0]+1,) + self.shape[1:]
        else:
            self.dtype = plane.dtype
            self.shape = (1,) + plane.shape
            self.sample_index = plane.sample_index
        self.planes.append(plane)

    def extend(self, other):
        """ Extend tiff array with the content of another.
        """
        list(map(self.append, other.planes))
              
    def get_voxel_sizes(self):
        """ Return ZYX voxel sizes in microns.
        """
        return self.planes[0].ifd.get_voxel_sizes()

    def get_pixel_sizes(self):
        """ Return YX pixel sizes in microns.
        """
        return self.planes[0].ifd.get_pixel_sizes()

    def get_time(self, index = 0):
        """ Return time parameter of a plane.
        """
        return self.planes[index].time

    @property
    def nbytes(self):
        return self.shape[0] * self.shape[1] * self.shape[2] * self.dtype.itemsize



"""A drop-in replacement for :pythonlib:`optparse` ("import iocbio.optparse_gui as optparse")

Provides an identical interface to
:pythonlib:`optparse`.OptionParser, in addition, it displays an
automatically generated `wx <http://www.wxpython.org>`_ dialog in order to enter the
options/args, instead of parsing command line arguments.

This code is based on `optparse_gui module
<http://code.google.com/p/optparse-gui/>`_. Unfortunately the owner of
optparse_gui did not respond to a request to join the optparse_gui
group. By now this module has become more complete with more features.

Module content
--------------
"""
#Author: Pearu Peterson
#Created: September 2009
from __future__ import division, print_function

__all__ = ['OptionParser', 'Option']

import os
import signal
import sys
import re
import shutil
import tempfile
import optparse
import wx
import subprocess as std_subprocess
if os.name=='nt':
    from . import killableprocess as subprocess
else:
    subprocess = std_subprocess

multiprocessing = None
try:
    import multiprocessing # Python <=2.5 users should install http://code.google.com/p/python-multiprocessing/
except ImportError:
    pass

import traceback
try:
    import queue as queue
except ImportError as e:
    import Queue as queue

from .utils import splitcommandline

#try:
#    import matplotlib
#except ImportError:
#    matplotlib = None
#if matplotlib is not None:
#    matplotlib.use('Agg') # to avoid crashes with other backends

__version__ = 0.9

debug = 1

bug_report_email = 'iocbio-bugs@googlegroups.com'
smtp_server = 'cens.ioc.ee'

bug_report_email = None # disables the Bug Report button

signal_map = {}
for n,v in list(signal.__dict__.items ()):
    if isinstance(v, int):
        signal_map[v] = n

def pretty_signal (s):
    r = signal_map.get(s)
    if r is None:
        r = signal_map.get(abs(s))
        if r is None:
            return str(s)
    return '%s[%s]' % (s,r)

def fix_float_string(s):
    if not s: return '0'
    if s[-1] in 'eE+-': return s + '0'
    return s

def fix_int_string (s):
    if s in ['+','-']: return s + '0'
    return s

def get_fixed_option_value (option, s):
    if option.type=='float': return fix_float_string(s)
    if option.type=='int': return fix_int_string(s)
    if option.type=='long': return fix_int_string(s)
    if option.type=='choice':
        choices = list(map(str, option.choices))
        try:
            return option.choices[choices.index(str(s))]
        except ValueError:
            pass
    return s

def os_kill(pid, s):
    if os.name=='nt':
        if s==signal.SIGTERM or s==signal.SIGINT:
            import win32api
            handle = win32api.OpenProcess(1, 0, pid)
            return (0 != win32api.TerminateProcess(handle, 127))
        if s==signal.SIGINT:
            import ctypes
            return (0 != ctypes.windll.kernel32.GenerateConsoleCtrlEvent (1, pid))
    return os.kill (pid, s)


class SysOutListener:
    debug_stream = '__stdout__'
    def __init__ (self, queue):
        self.queue = queue

    def write(self, string):
        if debug:
            stream = getattr (sys, self.debug_stream)
            if multiprocessing is not None or not string.startswith('\r'):
                stream.write(string)
            stream.flush()
        if multiprocessing is None:
            self.queue.put(string)
        else:
            if not self.queue._closed:
                self.queue.put(string)
        #wx.WakeUpIdle() # it does not seem to have effect, so using wx.Timer below

    def flush (self):
        if debug:
            stream = getattr (sys, self.debug_stream)
            stream.flush()

class SysErrListener (SysOutListener):
    debug_stream = '__stderr__'

    def write (self, string):
        if not self.queue._closed:
            self.queue.put('@STDERR START@')
        SysOutListener.write (self, string)
        if not self.queue._closed:
            self.queue.put('@STDERR END@')

class FunctionWrapper (object):

    def __init__ (self, func, log_queue = None):
        self.log_queue = log_queue
        self.func = func

    def __call__(self, *args, **kw):
        if self.log_queue is not None:
            saved_streams =  sys.stdout, sys.stderr
            sys.stdout = SysOutListener(self.log_queue)
            sys.stderr = SysErrListener(self.log_queue)
        
        status = 0
        try:
            self.func(*args, **kw)
        except:
            print("Exception in runner process:")
            print('-'*60)
            traceback.print_exc(file=sys.stdout)
            print('-'*60)        
            status = 1

        if self.log_queue is not None:
            sys.stdout, sys.stderr = saved_streams
        return status

class OptionValidator (wx.PyValidator):

    def __init__ (self, option):
        wx.PyValidator.__init__ (self)
        self.option = option
        self.Bind (wx.EVT_CHAR, self.OnChar)

    def Clone (self):
        return OptionValidator(self.option)
    
    def Validate (self, win):
        return True

    def OnChar(self, event):
        key = event.GetKeyCode()
        if key < wx.WXK_SPACE or key == wx.WXK_DELETE or key > 255:
            event.Skip ()
            return
        if self.option.type == 'choice':
            return

        textCtrl = self.GetWindow()
        i = textCtrl.GetInsertionPoint()
        char = chr(key)
        text = textCtrl.GetValue()
        new_text = get_fixed_option_value(self.option, text[:i] + char + text[i:])
        try:
            self.option.TYPE_CHECKER[self.option.type](self.option, self.option.dest, new_text)
        except optparse.OptionValueError as msg:
            if not wx.Validator_IsSilent():
                wx.Bell()
            print(msg, file=sys.stderr)
            return
        event.Skip()


class OptionPanel( wx.Panel ):
    
    def __init__(self, parent, option_list, main_frame):
        wx.Panel.__init__ (self, parent)

        sizer = wx.FlexGridSizer(cols=3, hgap=5, vgap=5)
        sizer.AddGrowableCol(1)

        destinations_list = []
        for option in option_list:

            help = option.help or 'specify %s option. Default: %%default.' % (option.get_opt_string())
            if '%default' in help:
                if option.default == optparse.NO_DEFAULT:
                    help = help.replace ('%default', 'none')
                else:
                    if option.action == 'store_false':
                        help = help.replace ('%default', str(not option.default if option.default is not None else option.default))
                    else:
                        help = help.replace ('%default', str(option.default))

            if option.action == 'store':
                label_txt = option.get_opt_string() + '='
                label = wx.StaticText(self, -1, label_txt )
                label.SetHelpText( help)
                sizer.Add (label, 0, wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL)

                if option.type == 'choice':
                    if option.default == optparse.NO_DEFAULT:
                        option.default = option.choices[0]
                    ctrl = wx.ComboBox(
                        self, -1, choices = list(map(str,option.choices)),
                        value = str(option.default),
                        style = wx.CB_DROPDOWN | wx.CB_READONLY | wx.CB_SORT 
                    )
                    sizer.Add (ctrl, 0, wx.EXPAND)
                    sizer.Add ((20,20),0)
                elif option.type in ['string', 'int', 'float']:
                    ctrl = wx.TextCtrl( self, -1, "", style=wx.TE_PROCESS_ENTER)
                    if option.default not in [optparse.NO_DEFAULT, None]:
                        ctrl.Value = str(option.default)
                    sizer.Add (ctrl, 0, wx.EXPAND)
                    sizer.Add ((20,20),0)
                elif option.type == 'multiline':
                    ctrl = wx.TextCtrl( self, -1, "", size=(300,100), 
                                        style = wx.TE_MULTILINE|wx.TE_PROCESS_ENTER )
                    if option.default not in [optparse.NO_DEFAULT, None]:
                        ctrl.Value = option.default
                    sizer.Add (ctrl, 0, wx.EXPAND)
                    sizer.Add ((20,20),0)
                elif option.type in ['file', 'directory']:
                    ctrl = wx.TextCtrl( self, -1, "")
                    if option.default not in [optparse.NO_DEFAULT, None]:
                        ctrl.Value = option.default
                    sizer.Add (ctrl, 0, wx.EXPAND)

                    browse = wx.Button( self, label='Browse..' )
                    browse.SetHelpText( 'Click to open %s browser' % (option.type) )
                    wx.EVT_BUTTON(self, browse.GetId(), main_frame.OnSelectPath)
                    main_frame.browse_option_map[browse.GetId()] = option, ctrl
                    sizer.Add(browse, 0, wx.ALIGN_CENTRE|wx.ALL, 1 )
                else:
                    raise NotImplementedError (repr(option.type))
                destinations_list.append(option.dest)

            elif option.action in ['store_true', 'store_false']:
                if option.dest in destinations_list:
                    continue
                ctrl = wx.CheckBox( self, -1, option.get_opt_string())
                #ctrl = wx.ToggleButton( self, -1, option.get_opt_string())
                if option.default not in [None, optparse.NO_DEFAULT]:
                    if option.action=='store_false':
                        ctrl.SetValue(not option.default)
                    else:
                        ctrl.SetValue(option.default)
                sizer.Add (ctrl, 0, wx.EXPAND)
                #sizer.Add ((20,20),0)
                sizer.Add (wx.StaticText(self,-1,help),0,wx.ALIGN_CENTER_VERTICAL)
                sizer.Add ((20,20),0)
                destinations_list.append(option.dest)
            elif option.action == 'help':
                ctrl = wx.Button( self, -1, 'Help')
                sizer.Add (ctrl, 0, wx.ALIGN_CENTER_VERTICAL|wx.ALIGN_RIGHT)
                #sizer.Add ((20,20),0)
                sizer.Add (wx.StaticText(self,-1,"print help message"),0,wx.ALIGN_CENTER_VERTICAL)
                sizer.Add ((20,20),0)
                self.Bind(wx.EVT_BUTTON, main_frame.OnHelp, id=ctrl.GetId())
            elif option.action == 'store_const':
                if option.dest in destinations_list:
                    continue
            else:
                raise NotImplementedError (repr(option.action))

            if option.type in option.TYPE_CHECKER:
                ctrl.SetValidator(OptionValidator(option))

            ctrl.SetHelpText( help )
            main_frame.option_controls[ option ] = ctrl


        self.SetSizer(sizer)
        self.Fit()
        #sizer.SetSizeHints(self)

class OptparseFrame (wx.Frame):

    def capture_std_streams(self):
        global multiprocessing
        if multiprocessing is None:
            self.log_queue = queue.Queue()
        else:
            try:
                self.log_queue = multiprocessing.Queue()
            except WindowsError as msg:
                print('multiprocessing.Queue raised WindowsError: %s' % (msg))
                print('Disableing multiprocessing')
                multiprocessing = None
                self.log_queue = None
                return
                self.log_queue = queue.Queue()
        self._saved_streams =  sys.stdout, sys.stderr
        sys.stdout = SysOutListener(self.log_queue)
        sys.stderr = SysErrListener(self.log_queue)

    def restore_std_streams (self):
        if self.log_queue is not None:
            sys.stdout, sys.stderr = self._saved_streams
            if multiprocessing is not None:
                self.log_queue.close()
            self.log_queue = None

    def __init__(self, option_parser):

        self.option_parser = option_parser
        self.option_parser.result = None 
        self.capture_std_streams()
        run_methods = getattr(option_parser, 'run_methods', ['subprocess', 'subcommand'])

        self.option_controls = {}
        self.browse_option_map = {}
        self.process_list = []

        wx.Frame.__init__(self, None, title="Optparse Dialog: %s" % (os.path.basename(sys.argv[0])))
        p = wx.Panel(self)

        provider = wx.SimpleHelpProvider()
        wx.HelpProvider_Set(provider)

        nb = wx.Notebook(p)
        
        option_panel = OptionPanel(nb, option_parser.option_list, self)
        option_panel.SetHelpText (option_parser.description or 'Main options')
        nb.AddPage(option_panel, 'Main Options')

        pages = []
        for group in option_parser.option_groups:
            if group in pages:
                continue
            option_panel = OptionPanel(nb, group.option_list, self)
            option_panel.SetHelpText (group.description or 'Group options')
            nb.AddPage(option_panel, group.title)
            pages.append(group)

        sizer_a = wx.BoxSizer (wx.VERTICAL)
        self.args_ctrl = args_ctrl = wx.TextCtrl(p, -1, '', size = ( -1, 80 ), 
                                                 style=wx.TE_MULTILINE | wx.TE_PROCESS_ENTER )
        args_ctrl.SetHelpText ('''\
Arguments can be either separated by a space or a newline. Arguments
that contain spaces must be entered like so: "arg with space"\
''')
        label = wx.StaticText(p, -1, 'Arguments:' )
        label.SetHelpText( 'Enter arguments here' )
        sizer_a.Add (label, 0, wx.EXPAND)
        sizer_a.Add (args_ctrl, 0, wx.EXPAND)
        args_ctrl.Fit ()

        sizer_b = wx.BoxSizer (wx.HORIZONTAL)


        cancel_b = wx.Button(p, -1, 'Cancel')
        cancel_b.SetHelpText('Cancel the program without saving')

        if debug>1:
            test_b = wx.Button(p, -1, 'Test')
            test_b.SetHelpText('Test options')

        run_b = None
        if option_parser.runner is not None or 1:
            self.run_b = run_b = wx.Button(p, -1, 'Run')
            run_b.SetHelpText('Run script with specified options and arguments.')

        batch_b = None
        if 0:
            self.batch_b = batch_b = wx.Button(p, -1, 'Batch')
            batch_b.SetHelpText('Run script with specified options and arguments in batch mode and exit.')

        self.exit_b = exit_b = wx.Button(p, -1, 'Exit')
        exit_b.SetHelpText('Exit without running the script. Any running process will be terminated.')

        send_b = None
        if bug_report_email:
            send_b = wx.Button (p, -1, 'Send Bug Report')
            send_b.SetHelpText ('Send bug report to %r. The bug report will include some environment data, used options and arguments, the content of the log window. If you want to add some notes then enter them to log window before pressing the Send Bug Report button.' % (bug_report_email))

        runner_mth = None
        if multiprocessing is None or option_parser.runner is None:
            self.run_method = 'subcommand'
        else:
            self.run_method = run_methods[0]
            if len (run_methods)>1:
                runner_mth = wx.ComboBox(
                    p, -1, choices = run_methods,
                    value = run_methods[0],
                    style = wx.CB_DROPDOWN | wx.CB_READONLY | wx.CB_SORT 
                    )
                runner_mth.SetHelpText ('Specify how to run application. Selecting subprocess means using multiprocessing.Process. Selecting subcommand means running the application with --no-gui option and using specified options. If application uses wx then one should select subcommand.')

        sizer_b.Add(cancel_b, 0, wx.ALL, 2)
        if debug>1:
            sizer_b.Add(test_b, 0, wx.ALL, 2)
        if run_b is not None:
            sizer_b.Add(run_b, 0, wx.ALL, 2)
            if runner_mth is not None:
                sizer_b.Add(runner_mth, 0, wx.ALL, 2)
        if batch_b is not None:
            sizer_b.Add(batch_b, 0, wx.ALL, 2)
        sizer_b.Add(exit_b, 0, wx.ALL, 2)
        if send_b is not None:
            sizer_b.Add(send_b, 0, wx.ALL, 2)

        self.log_window = log_window = wx.TextCtrl (p, -1, '', style = wx.TE_MULTILINE,
                                                    size = ( -1, 120 ))
        log_window.SetFont(wx.Font(8, wx.MODERN, wx.NORMAL, wx.NORMAL, False))
        log_window.SetHelpText ('Shows the standard output and error messages of a running process.')

        if wx.Platform != "__WXMSW__" or 1:
            help_b = wx.ContextHelpButton(p)
            sizer_b.Add((20,20),1)
            sizer_b.Add(help_b, 0, wx.RIGHT, 2)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(nb, 0, wx.EXPAND)
        sizer.Add(sizer_a, 0, wx.EXPAND, 5 )
        sizer.Add(sizer_b, 0, wx.EXPAND|wx.BOTTOM, 5)
        sizer.Add(log_window, 1, wx.EXPAND)

        p.SetSizer(sizer)
        p.Fit ()
        sizer.SetSizeHints(p)

        self.Fit ()

        self.timer = wx.Timer(self)


        self.Bind(wx.EVT_CLOSE, self.OnCloseWindow)
        self.Bind(wx.EVT_BUTTON, self.OnCancel, id=cancel_b.GetId())
        if run_b is not None:
            self.Bind(wx.EVT_BUTTON, self.OnRun, id=run_b.GetId())
            if runner_mth is not None:
                self.Bind(wx.EVT_COMBOBOX, self.OnSelectRunnerMethod, id=runner_mth.GetId())
        if batch_b is not None:
            self.Bind(wx.EVT_BUTTON, self.OnBatch, id=batch_b.GetId())
        if debug>1:
            self.Bind(wx.EVT_BUTTON, self.OnTest, id=test_b.GetId())
        self.Bind(wx.EVT_BUTTON, self.OnExit, id=exit_b.GetId())
        if send_b is not None:
            self.Bind(wx.EVT_BUTTON, self.OnSend, id=send_b.GetId())
        self.Bind(wx.EVT_IDLE, self.OnUpdateLogWindow)
        self.Bind(wx.EVT_TIMER, self.OnUpdateLogWindow, self.timer)

        self.previous_position = 0
        self._running_count = 0

    def OnSend(self, event):
        msg = ''

        msg += '-'*40 + '\n'
        msg += 'Environment\n'
        msg += '='*40 + '\n'
        msg += 'os.getcwd(): %r\n' % (os.getcwd())
        msg += 'sys.argv: ' + repr(sys.argv) + '\n'
        msg += 'sys.version: ' + repr(sys.version) + '\n'
        msg += 'sys.prefix: ' + repr(sys.prefix) + '\n'
        user = None
        for k in ['USER', 'LOGNAME', 'HOME', 'PYTHONPATH', 'PYTHONSTARTUP', 'LD_LIBRARY_PATH', 'PATH',
                  'LTSP_CLIENT','SSH_CONNECTION']:
            v = os.environ.get (k, None)
            if v is not None:
                msg += '%s: %r\n' % (k,v)

                if user is None and k in ['USER', 'LOGNAME']:
                    user = v
        msg += 'Installation path: %r\n' % (os.path.dirname (__file__))
        try:
            from . import version
            msg += 'Version: %s\n' % (version.version)
        except Exception as msg:
            msg += '%s\n' % (msg)
            pass
        msg += '-'*40 + '\n'

        msg += '\n'
        msg += '-'*40 + '\n'
        msg += 'Options\n'
        msg += '='*40 + '\n'    
        try:
            values, args = self.option_parser.parse_options_args
            self.set_result()
            values, args = self.option_parser.get_result (values)
            msg += 'args=%r\n' % (args)
            for option in self.option_parser._get_all_options():
                if option.dest:
                    value = getattr(values, option.dest, None)
                    if value is not None:
                        msg += '%s[dest=%s]: %r\n' % (option.get_opt_string(), option.dest, value)
        except Exception as m:
            msg += 'Failed to get option results: %s\n' % (m)
        msg += '-'*40 + '\n'

        msg += '\n'
        msg += '-'*40 + '\n'
        msg += 'Content of log window\n'
        msg += '='*40 + '\n'
        msg += self.log_window.GetValue() + '\n'
        msg += '-'*40 + '\n'

        message = msg

        user_email = user + "@" + smtp_server
        import smtplib
        from email.mime.text import MIMEText
        msg = MIMEText(msg)
        msg['Subject'] = '[IOCBio bug report] %s' % (sys.argv[0])
        msg['From'] = user_email
        msg['To'] = bug_report_email

        try:
            s = smtplib.SMTP()
            s.connect(smtp_server)
            # Send the email - real from, real to,
            # extra headers and content ...
            s.sendmail(user_email, bug_report_email, msg.as_string())
            s.close()
            print('Bug report succesfully sent to %r' % (bug_report_email))
        except Exception as msg:
            print('Failed to send bug report: %s' % (msg))
            f = open('iocbio_bug_report.txt', 'w')
            f.write (message)
            f.close()
            print('Please find the file "iocbio_bug_report.txt" in the '
                  'current working directory and send it to %r using your '
                  'favorite E-mail program.' % bug_report_email)

    def OnSelectRunnerMethod(self, event):
        self.run_method = event.GetString()

    def OnUpdateLogWindow(self, event):
        if self.log_queue is not None and not self.log_queue.empty():
            log_window = self.log_window
            is_stderr_message = None
            while 1:
                try:
                    v = self.log_queue.get_nowait()
                except queue.Empty:
                    break
                if v=='@STDERR START@':
                    is_stderr_message = True
                elif v=='@STDERR END@':
                    is_stderr_message = False
                else:
                    # todo: use is_stderr_message to switch styles
                    for line in v.splitlines(True):
                        if line.startswith('\r'):
                            log_window.Remove(self.previous_position, log_window.GetLastPosition())
                            line = line[1:]
                        self.previous_position = log_window.GetInsertionPoint()
                        log_window.AppendText(line)
            self.log_window.Refresh()

        if self.process_list:
            for index, (process, method) in enumerate(self.process_list):
                if method=='subprocess':
                    if not process.is_alive():
                        self._cleanup_process(process, method)
                        del self.process_list[index]
                elif method=='subcommand':
                    if process.poll() is not None:
                        self._cleanup_process(process, method)
                        del self.process_list[index]
                else:
                    raise NotImplementedError (repr(method))
            self.cleanup_runner()

    def OnHelp (self, event):
        self.option_parser.print_help()

    def OnTest (self, event):
        values, args = self.option_parser.parse_options_args
        self.set_result()
        new_values, new_args = self.option_parser.get_result (values)
        print(new_values, new_args)

    def OnBatch (self, event):
        self.interrupt_runner()
        self.set_result()
        self.restore_std_streams()
        self.Close(True)

    def OnCancel (self, event):
        while self.process_list:
            self.interrupt_runner()
        self.result = None
        self.restore_std_streams()
        self.Close(True)
        sys.exit(0)

    def _start_process (self, method, show_info=[True]):
        self.set_result()
        values, args = self.option_parser.parse_options_args
        new_options, new_args = self.option_parser.get_result (values)
        if method=='subprocess':
            process = multiprocessing.Process(target=FunctionWrapper(self.option_parser.runner,
                                                                     self.log_queue), 
                                              args=(self.option_parser, new_options, new_args))
            process.start ()
            return process
        elif method=='subcommand':
            if multiprocessing is None:
                print('Using multiprocessing package is disabled.')
                if sys.version[:3]<='2.5':
                    print('Python 2.4, 2.5 users should install multiprocessing package from http://code.google.com/p/python-multiprocessing/')
            if show_info[0]:
                print('Live session output is visible only in terminal.')
                print('Wait for the runner process to finish..')
                show_info[0] = False
            self.option_parser.save_options(new_options, new_args)
            cmd_lst = []
            cmd_prefix = getattr(new_options, 'runner_subcommand', sys.executable)
            if cmd_prefix:
                cmd_lst.append(cmd_prefix)
            cmd_lst.append (sys.argv[0])
            cmd_lst.append ('--no-gui')
            cmd = ' '.join(cmd_lst)
            print('Executing command:', cmd)
            process = subprocess.Popen (cmd, shell=True,
                                        #stdout=std_subprocess.PIPE,
                                        stderr=std_subprocess.STDOUT)
            return process
        else:
            raise NotImplementedError (repr(method))

    def _cleanup_process (self, process, method):
        if process is None:
            return
        if method=='subprocess':
            if process.exitcode is None:
                pid = process.pid
                print('Runner %s (PID=%s) still running, trying to terminate' % (method, pid))
                process.terminate()
                process.join()
            print('Runner %s has finished with exitcode=%s' % (method,  pretty_signal(process.exitcode)))
        elif method=='subcommand':
            s = process.wait()
            if os.name=='nt':
                print('Runner %s has finished with returncode=%s' % (method,  pretty_signal (s)))
            else:
                pid = process.pid
                print('Runner %s (PID=%s) has finished with returncode=%s' % (method, pid, pretty_signal(s)))
        else:
            raise NotImplementedError (repr(method))

    def _interrupt_process (self, process, method):
        if method=='subprocess':
            if process.is_alive():
                pid = process.pid
                os_kill(pid, signal.SIGINT)
                print()
                print('SIGINT signal has been sent to runner %s (PID=%s)' % (method, pid))
        elif method=='subcommand':
            if process.poll () is None:
                if os.name=='nt':
                    process.kill(True, s=signal.SIGINT)
                    print()
                    print('SIGINT signal has been sent to runner %s' % (method))
                else:
                    pid = process.pid
                    os_kill(pid, signal.SIGINT)
                    print()
                    print('SIGINT signal has been sent to runner %s (PID=%s)' % (method, pid))
        else:
            raise NotImplementedError (repr(method))

    def _terminate_process (self, process, method):
        if method=='subprocess':
            if prosess.is_alive():
                print()
                print('Terminating runner %s' % (method))
                process.terminate()
        elif method=='subcommand':
            if process.poll() is None:
                if sys.version[:3]>='2.6':
                    print()
                    print('Terminating runner %s' % (method))
                    process.terminate ()
                else:
                    if os.name=='nt':
                        process.kill (True, s = signal.SIGTERM)
                    else:
                        os_kill (process.pid, signal.SIGTERM)
                    print()
                    print('SIGTERM signal has been sent to runner %s' % (method))
        else:
            raise NotImplementedError (repr(method))

    def start_runner (self):
        if not self.process_list:
            self.exit_b.SetLabel('Stop')
            self.timer.Start(100)
        self.option_parser.run_method = self.run_method
        process = self._start_process(self.run_method)
        pid = process.pid
        print()
        print('Runner %s (PID=%s) has been started' % (self.run_method, pid))
        self.process_list.append((process, self.run_method))

    def cleanup_runner (self):
        if not self.process_list:
            self.timer.Stop()
            self.exit_b.SetLabel('Exit')

    def interrupt_runner (self):
        if self.process_list:
            process, run_method = self.process_list[-1]
            del self.process_list[-1]
            self._interrupt_process(process, run_method)
            self._cleanup_process(process, run_method)
        self.cleanup_runner()

    def terminate_runner(self):
        if self.process_list:
            process, run_method = self.process_list[-1]
            del self.process_list[-1]
            self._terminate_process(process, run_method)
            self._cleanup_process(process, run_method)
        self.cleanup_runner()

    def OnExit(self, event):
        if self.exit_b.GetLabel()=='Stop':
            self.interrupt_runner()
            return

        while self.process_list:
            self.terminate_runner()

        try:
            values, args = self.option_parser.parse_options_args
            self.set_result()
            values, args = self.option_parser.get_result(values)
            #self.option_parser.save_options(values, args)
        except optparse.OptionValueError as msg:
            print('Ignoring %s' % (msg))
            pass

        self.restore_std_streams()
        self.Close(True)
        sys.exit(0)

    def OnRun(self, event):
        self.start_runner()

    def OnCloseWindow (self, event):
        while self.process_list:
            self.terminate_runner()
        self.restore_std_streams()
        self.result = None
        self.Destroy()

    def OnSelectPath(self, event):
        option, ctrl = self.browse_option_map[event.GetId()]
        path = os.path.abspath(ctrl.Value)
        if option.type == 'file':
            if os.path.isdir (path):
                default_file = ''
                default_dir = path
            else:
                default_file = path
                default_dir =os.path.dirname(path)
            dlg = wx.FileDialog(self,
                                message = 'Select file for %s' % (option.get_opt_string()),
                                defaultDir = default_dir, defaultFile=default_file)
        elif option.type == 'directory':
            if os.path.isfile(path):
                path = os.path.dirname (path)
            dlg = wx.DirDialog(self,
                               message = 'Select directory for %s' % (option.get_opt_string()),
                               defaultPath = path)
        else:
            raise NotImplementedError(repr(option.type))
        if dlg.ShowModal() != wx.ID_OK:
            return
        cwd = os.path.abspath(os.getcwd())
        value = dlg.GetPath()
        if value.startswith(cwd):
            value = value[len(cwd)+1:]
        ctrl.Value = value

    def set_result(self):
        option_values = {}
        for option, ctrl in self.option_controls.items():
            value = getattr(ctrl,'Value',None)
            if value != '' and value is not None:
                value = get_fixed_option_value(option, value)
                option_values[option] = value
            elif value is not None:
                option_values[option] = None
        args_buff = str(self.args_ctrl.GetValue())
        args =  splitcommandline(args_buff)
        self.option_parser.result = option_values, args

    def put_result(self):
        option_values, args = self.option_parser.result
        for option, ctrl in self.option_controls.items():
            value = option_values.get(option)
            if value is not None:
                ctrl.SetValue(value)
        self.args_ctrl.SetValue(' '.join(args))

class UserCancelledError( Exception ):
    pass

def check_file(option, opt, value):
    try:
        value = str(value)
    except ValueError:
        raise OptionValueError(
            _("option %s: invalid %s value: %r") % (opt, what, value))
    #if value and not os.path.isfile(value):
    #    print 'Warning: path %r is not a file' % (value)
    return value

def check_directory(option, opt, value):
    try:
        value = str(value)
    except ValueError:
        raise OptionValueError(
            _("option %s: invalid %s value: %r") % (opt, what, value))
    #if value and not os.path.isdir(value):
    #    print 'Warning: path %r is not a directory' % (value)
    return value

class Option(optparse.Option):
    """Extends optparse.Option with file, directory and multiline types.
    """
    _SUPER = optparse.Option
    TYPES = _SUPER.TYPES + ('file', 'directory', 'multiline')
    TYPE_CHECKER = _SUPER.TYPE_CHECKER.copy()
    TYPE_CHECKER.update (file=check_file, directory=check_directory)


class OptionParser( optparse.OptionParser ):
    """Extends optparse.OptionParser with GUI support.
    """
    _SUPER = optparse.OptionParser
    """Holds base class.
    """
    
    def __init__( self, *args, **kwargs ):
        if 'option_class' not in kwargs:
            kwargs['option_class'] = Option
        self.runner = None
        self._SUPER.__init__( self, *args, **kwargs )

    def save_option_value (self, dest, value):
        return self._set_dest_value (dest, value)

    def _set_dest_value(self, dest, value, old_cwd = None):
        if value is None:
            return
        option = ([option for option in self._get_all_options () if option.dest == dest] + [None])[0]
        if option is None:
            print('Could not find option with dest=%r for setting %r (dest not specified for option).' % (dest, value))
        else:
            if option.type in ['file', 'directory']:
                if old_cwd is not None:
                    value = os.path.join(old_cwd, value)
                else:
                    value = os.path.abspath(value)
                cwd = os.getcwd()
                if value.startswith(cwd):
                    value = value[len(cwd)+1:]
            if value == 'none':
                value = None
            else:
                try:
                    option.check_value(option.dest, str(value))
                except optparse.OptionValueError as msg:
                    print('_set_dest_value: ignoring %s' % msg)
                    return
                option.default = value
        if value is None:
            try:
                del self.defaults[dest]
            except KeyError:
                pass
        else:
            self.defaults[dest] = value

    def get_history_file(self):
        import hashlib
        script_history = os.path.join(os.environ.get('HOME',''), '.optparse_history', 
                                      os.path.basename(sys.argv[0]) + hashlib.md5(os.path.abspath(sys.argv[0])).hexdigest())        
        return script_history


    def save_options (self, values, args):
        script_history = self.get_history_file()
        if debug>1:
            print('Saving options to', script_history)
        cwd = os.path.abspath(os.getcwd())
        dirname = os.path.dirname(script_history)
        if not os.path.isdir (dirname):
            os.makedirs(dirname)

        tmp_file = tempfile.mktemp()
            
        f = open(tmp_file, 'w')
        f.write ('#cwd:%r\n' % (cwd))
        f.write ('#args:%r\n' % (args,))
        for option in self._get_all_options():
            if option.dest:
                value = getattr(values, option.dest, None)
                if value is not None:
                    f.write('%s: %r\n' % (option.dest, value))
        f.close()        

        shutil.move(tmp_file, script_history)
        
    def load_options(self):
        script_history = self.get_history_file ()
        if debug>1:
            print('Loading options from',script_history)
        h_args = None
        h_cwd = None
        cwd = os.path.abspath(os.getcwd())
        if os.path.isfile(script_history):
            f = open (script_history)
            for line in f.readlines():
                try:
                    dest, value = line.split(':', 1)
                    value = eval(value)
                except Exception as msg:
                    print('optparse_gui.load_options: failed parsing options file, line=%r: %s' % (line, msg))
                    continue
                if dest=='#args':
                    h_args = value
                elif dest=='#cwd':
                    h_cwd = value
                else:
                    self._set_dest_value(dest, value, h_cwd)
            f.close()        
        return h_args

    def parse_args( self, args = None, values = None ):
        '''Parse the command-line options.
        '''
        # load options history
        h_args = self.load_options()

        no_gui = '--no-gui' in sys.argv
        if no_gui:
            sys.argv.remove('--no-gui')
        # preprocess command line arguments and set to defaults
        pp_option_values, pp_args = self._SUPER.parse_args(self, args, values)

        if no_gui:
            self.save_options(pp_option_values, pp_args)
            return pp_option_values, pp_args

        for dest, value in pp_option_values.__dict__.items():
            self._set_dest_value(dest, value)

        self.parse_options_args = (values, args)

        app = wx.App(redirect=False)
        try:
            dlg = OptparseFrame(self)
        except Exception as msg:
            traceback.print_exc(file=sys.stdout)
            raise
        if pp_args:
            dlg.args_ctrl.Value = ' '.join(pp_args)
        elif h_args is not None:
            dlg.args_ctrl.Value = ' '.join(h_args)
        dlg.Show (True)
        app.MainLoop()

        if self.result is None:
            print('User has cancelled, exiting.')
            sys.exit(0)
        
        values, args = self.get_result(values)
        self.save_options(values, args)

        return values, args

    def error( self, msg ):
        print("AN ERROR OCCURRED WITH A MESSAGE: %s" % (msg))
        #app = wx.GetApp()
        #print app, msg
        #if app is None:    
        #    app = wx.App( False )
        #wx.MessageDialog( None, msg, 'Error!', wx.ICON_ERROR ).ShowModal()
        return self._SUPER.error( self, msg )

    def get_result(self, values):
        if values is None:
            values = self.get_default_values()
            
        option_values, args = self.result
         
        for option, value in option_values.items():
            if option.action=='store_true':
                if isinstance(value, bool):
                    setattr( values, option.dest, value )
                continue
            if option.action=='store_false':
                if isinstance(value, bool):
                    setattr( values, option.dest, not value )
                continue
            
            if option.takes_value() is False:
                value = None

            option.process( option, value, values, self )
        return values, args

################################################################################

def sample_parse_args():
    usage = "usage: %prog [options] args"
    if 1 == len( sys.argv ):
        option_parser_class = OptionParser
    else:
        option_parser_class = optparse.OptionParser
        
    parser = option_parser_class( usage = usage, version='0.1' )
    parser.add_option("-f", "--file", dest="filename", default = r'c:\1.txt',
                      help="read data from FILENAME")
    parser.add_option("-t", "--text", dest="text", default = r'c:\1.txt',
                      help="MULTILINE text field")
    parser.add_option("-a", "--action", dest="action",
                      choices = ['delete', 'copy', 'move'],
                      help="Which action do you wish to take?!")
    parser.add_option("-n", "--number", dest="number", default = 23,
                      type = 'int',
                      help="Just a number")
    parser.add_option("-v", "--verbose",
                      action="store_true", dest="verbose",
                      help = 'To be or not to be? ( verbose )' )
    
    (options, args) = parser.parse_args()
    return options, args

def sample_parse_args_issue1():
    usage = "usage: %prog [options] args"
    option_parser_class = OptionParser
        
    parser = option_parser_class( usage = usage, version='0.1' )
    parser.add_option("-f", "--file", dest="filename", default = r'c:\1.txt',
                      type = 'file',
                      help="read data from FILENAME")
    parser.add_option("-t", "--text", dest="text", default = r'c:\1.txt',
                      type = 'multiline',
                      help="MULTILINE text field")
    parser.add_option("-a", "--action", dest="action",
                      choices = ['delete', 'copy', 'move'],
                      help="Which action do you wish to take?!")
    parser.add_option("-n", "--number", dest="number", default = 23,
                      type = 'int',
                      help="Just a number")
    parser.add_option("-v", "--verbose",
                      action="store_true", dest="verbose",
                      help = 'To be or not to be? ( verbose )' )

    group = optparse.OptionGroup(parser, "Dangerous Options",
                        "Caution: use these options at your own risk.  "
                        "It is believed that some of them bite.")
    group.add_option("-g", action="store_true", help="Group option.",
                     )
    parser.add_option_group(group)

    (options, args) = parser.parse_args()
    return options, args

def main():
    options, args = sample_parse_args_issue1()
    print('args: %s' % repr( args ))
    print('options: %s' % repr( options ))
    
if '__main__' == __name__:
    main()




__all__ = ['TiffFiles']

import time

from .tiff_file import TiffFile
from .tiff_array import TiffArray
from .tiff_sample_plane import TiffSamplePlane, TiffSamplePlaneLazy
from .tiff_base import TiffBase

class TiffFiles(TiffBase):
    """Represent a collection of TIFF files as a single TIFF source object.

    See also
    --------
    TiffFile, TiffChannelsAndFiles
    """

    def __init__(self, files, time_map = {}, verbose = False, local_cache = None):
        """
        Parameters
        ----------
        files : list
          A list of TIFF file names.
        time_map : dict
          A dictionary of TIFF file names and a list of time values
          corresponding to image file directories (IFDs) in the
          corresponding TIFF files.
        verbose : bool
        local_cache : {None, str}
          Specify path to local cache. Local cache will be used to
          temporarily store files from external devises such as NFS.
        """
        self.verbose = verbose
        self.files = files
        self.tiff_files = {}
        self.time_map = time_map
        self.local_cache = local_cache

    def get_tiff_file(self, filename, use_memmap=True):
        tiff = self.tiff_files.get(filename)
        if tiff is None:
            tiff = TiffFile(filename, verbose=self.verbose, local_cache = self.local_cache, use_memmap=use_memmap)
            #tiff.set_time(self.time_map.get(filename))
            self.tiff_files[filename] = tiff
        return tiff

    def get_tiff_array(self, sample_index = 0, subfile_type=0, assume_one_image_per_file=False, use_memmap=True):
        """ Return an array of images.

        Parameters
        ----------
        sample_index : int
          Specify sample within a pixel.
        subfile_type : int
          Specify TIFF NewSubfileType used for collecting sample images.
        assume_one_image_per_file : bool
          When True then it is assumed that each TIFF file contains
          exactly one image and all images have the same parameters.
          This knowledge speeds up tiff_array construction as only the
          first TIFF file is opened for reading image parameters. The
          other TIFF files are opened only when particular images are
          accessed.
        use_memap : bool
          When True then image data is read in using numpy.memmap.

        Returns
        -------
        tiff_array : TiffArray
          Array of sample images. The array has rank equal to 3.
        """
        start = time.time ()
        planes = []
        
        if assume_one_image_per_file:
            for index, filename in enumerate (self.files):
                time_lst = self.time_map.get(filename)

                if index==0:
                    tiff = self.get_tiff_file(filename, use_memmap=use_memmap)
                    assert len (tiff.IFD)==1,repr(len (tiff.IFD))
                    ifd = tiff.IFD[0]
                    assert ifd.get_value('NewSubfileType', subfile_type)==subfile_type
                    plane = TiffSamplePlane(ifd, sample_index=sample_index)
                else:
                    def tiff_file_getter(parent=self, filename=filename):
                        tiff = parent.get_tiff_file(filename, use_memmap=use_memmap)
                        return tiff
                    plane = TiffSamplePlaneLazy(tiff_file_getter)
                    plane.copy_attrs(planes[0])

                if time_lst is not None:
                    assert len (time_lst)==1,repr(len(time_lst))
                    plane.set_time(time_lst[0])
                planes.append(plane)                    
        else:
            for filename in self.files:
                tiff = self.get_tiff_file(filename, use_memmap=use_memmap)
                time_lst = self.time_map.get(filename)
                index = 0
                for ifd in tiff.IFD:
                    if ifd.get_value('NewSubfileType', subfile_type) != subfile_type:
                        continue
                    plane = TiffSamplePlane(ifd, sample_index=sample_index)
                    if time_lst is not None:
                        plane.set_time(time_lst[index])
                    planes.append(plane)
                    index += 1

        tiff_array = TiffArray(planes)
        if self.verbose:
            print('%s.get_tiff_array: took %ss' % (self.__class__.__name__, time.time ()-start))
        return tiff_array

    def close (self):
        for tiff in list(self.tiff_files.values()):
            tiff.close()
        self.tiff_files.clear()

    __del__ = close


    def get_info(self, use_memmap=True):
        filename = self.files[0]
        tiff = self.get_tiff_file(filename, use_memmap=True)
        return tiff.get_info()



"""
lsm - implements TIFF CZ_LSM support.
"""
# Author: Pearu Peterson
# Created: April 2010

__all__ = ['scaninfo', 'timestamps', 'eventlist','channelcolors',
           'channelwavelength','inputlut','outputlut',
           'vectoroverlay','roi','bleachroi','meanofroisoverlay',
           'topoisolineoverlay','topoprofileoverlay','linescanoverlay',
           'channelfactors', 'channeldatatypes',
           'lsmblock', 'lsminfo'
           ]

import sys
import numpy

CZ_LSMInfo_tag = 0x866C
tiff_module_dict = None
verbose_lsm_memory_usage = False

def IFDEntry_lsm_str_hook(entry):
    l = []
    for name in entry.value.dtype.names:
        func = CZ_LSMOffsetField_readers.get(name)
        if func is not None:
            l.append('\n  %s->%s' % (name, func(entry, debug=False)))
        else:
            v = entry.value[name]
            l.append('\n  %s=%s' % (name, v))
    return 'IFDEntry(tag=%s, value=%s(%s))' % (entry.tag_name, entry.type_name, ', '.join (l))

def IFDEntry_lsm_init_hook(ifdentry):
    """Make tiff.IFDENTRYEntry CZ_LSM aware.
    """
    global tiff_module_dict
    if ifdentry.tag==CZ_LSMInfo_tag:
        # replace type,count=(BYTE,500) with (CZ_LSMInfo, 1)
        reserved_bytes = (ifdentry.count - CZ_LSMInfo_dtype_fields_size)
        CZ_LSMInfo_type = (CZ_LSMInfo_tag, reserved_bytes)
        if CZ_LSMInfo_type not in tiff_module_dict['type2dtype']:
            dtype = numpy.dtype(CZ_LSMInfo_dtype_fields + [('Reserved', numpy.dtype('(%s,)u4' % (reserved_bytes//4)))])
            CZ_LSMInfo_type_name = 'CZ_LSMInfo%s' % (reserved_bytes)
            CZ_LSMInfo_type_size = dtype.itemsize
            tiff_module_dict['type2dtype'][CZ_LSMInfo_type] = dtype
            tiff_module_dict['type2name'][CZ_LSMInfo_type] = CZ_LSMInfo_type_name
            tiff_module_dict['type2bytes'][CZ_LSMInfo_type] = CZ_LSMInfo_type_size
            tiff_module_dict['tag_name2value'][CZ_LSMInfo_type_name] = CZ_LSMInfo_tag
        #assert ifdentry.count==CZ_LSMInfo_type_size,`ifdentry.count,CZ_LSMInfo_type_size`
        ifdentry.type = CZ_LSMInfo_type
        ifdentry.count = 1
        ifdentry.tiff.is_lsm = True
        ifdentry.str_hook = IFDEntry_lsm_str_hook
    else:
        if not hasattr(ifdentry.tiff, 'is_lsm'):
            ifdentry.tiff.is_lsm = False

def IFDEntry_lsm_finalize_hook(ifdentry):
    if ifdentry.tag==CZ_LSMInfo_tag:
        blockstart, blockend = None, None
        for name in ifdentry.value.dtype.names:
            func = CZ_LSMOffsetField_readers.get(name)
            if func is not None:
                s = func(ifdentry, debug=False)
                if s is not None:
                    start = s.offset
                    end = start + s.get_size()
                    if blockstart is None:
                        blockstart, blockend = start, end
                    else:
                        blockstart, blockend = min(blockstart,start), max(blockend,end)
                    if verbose_lsm_memory_usage:
                        ifdentry.memory_usage.append((start, end,
                                                      ifdentry.tag_name+' '+name[6:]))
        if verbose_lsm_memory_usage:
            for offset in ifdentry.value['Reserved'][0]:
                if offset:
                    ifdentry.memory_usage.append((offset, offset, 'start of a unknown reserved field'))
        else:
            ifdentry.memory_usage.append((blockstart, blockend, 'lsmblock'))
        ifdentry.block_range = (blockstart, blockend)
        ifdentry.tiff.lsminfo = scaninfo(ifdentry, debug=False)
        ifdentry.tiff.lsmblock = lsmblock(ifdentry, debug=False)
        ifdentry.tiff.lsmentry = ifdentry.value

def register(tiff_dict):
    """Register CZ_LSM support in tiff module.
    """
    global tiff_module_dict
    tiff_module_dict = tiff_dict
    tiff_dict['IFDEntry_init_hooks'].append(IFDEntry_lsm_init_hook)
    tiff_dict['IFDEntry_finalize_hooks'].append(IFDEntry_lsm_finalize_hook)

def lsminfo(ifdentry, new_lsmblock_start=None):
    if ifdentry.tag!=CZ_LSMInfo_tag:
        return
    target = ifdentry.value.copy()
    old_lsmblock_start = ifdentry.block_range[0]
    if new_lsmblock_start is None:
        new_lsmblock_start = old_lsmblock_start
    offset_diff = new_lsmblock_start - old_lsmblock_start
    i = 0
    for name in ifdentry.value.dtype.names:
        dt = ifdentry.value.dtype[name]
        value_ref = target[name]
        value = value_ref[0]
        if name.startswith('Offset') and value!=0:
            #print 'changing',name,'from',value,'to',
            value_ref += offset_diff
            #print value_ref[0]
        elif name=='Reserved':
            # assuming that unknown values in Reserved field are offsets:
            for i in range(len(value)):
                if value[i]!=0:
                    if value[i]>=ifdentry.block_range[0] and value[i]<=ifdentry.block_range[1]:
                        #print 'chaning Reserved[%s] from %s to' % (i, value[i]),
                        value[i] += offset_diff
                        #print value[i]
                    else:
                        sys.stderr.write('Reserved[%s]=%s is out of block range %s. lsminfo might not be correct.' % (i,value[i], ifdentry.block_range))
        i += dt.itemsize    

    return target

class LSMBlock:
    
    def __init__(self, ifdentry):
        self.ifdentry = ifdentry
        self._offset = None

    @property
    def offset(self):
        if self._offset is None:
            self._offset = self.ifdentry.block_range[0]
        return self._offset

    def get_size(self):
        return self.ifdentry.block_range[1] - self.offset

    def __str__(self):
        return '%s(offset=%s, size=%s, end=%s)' % (self.__class__.__name__, self.offset, self.get_size(), self.offset+self.get_size())

    def get_data(self, new_offset):
        raise NotImplementedError(repr(new_offset))

    def toarray(self, target=None, new_offset=None):
        sz = self.get_size()
        if target is None:
            target = numpy.zeros((sz,), dtype = numpy.ubyte)
        if new_offset is None:
            new_offset = self.offset
        offset_diff = new_offset - self.offset
        dtype = target.dtype
        target[:sz] = self.ifdentry.tiff.data[self.offset:self.offset + sz]
        return target


def lsmblock(ifdentry, debug=True):
    if ifdentry.tag!=CZ_LSMInfo_tag:
        return
    r = LSMBlock(ifdentry)
    if debug:
        arr = r.toarray()
        offset = r.offset
        arr2 = ifdentry.tiff.data[offset:offset+arr.nbytes]
        assert (arr==arr2).all()
    return r

class ScanInfoEntry:
    """ Holds scan information entry data structure.
    """
    def __init__(self, entry, type_name, label, data):
        self._offset = None
        self.record = (entry, type_name, label, data)
        self.footer = None
        if type_name == 'ASCII':
            self.type = 2
            self.header = numpy.array([entry, 2, len(data)+1], dtype=numpy.uint32).view(dtype=numpy.ubyte)
        elif type_name == 'LONG':
            self.type = 4
            self.header = numpy.array([entry, 4, 4], dtype=numpy.uint32).view(dtype=numpy.ubyte)
        elif type_name == 'DOUBLE':
            self.type = 5
            self.header = numpy.array([entry, 5, 8], dtype=numpy.uint32).view(dtype=numpy.ubyte)
        elif type_name == 'SUBBLOCK':
            self.type = 0
            self.header = numpy.array([entry, 0, 0], dtype=numpy.uint32).view(dtype=numpy.ubyte)
        else:
            raise NotImplementedError (repr(self.record))

    def __repr__ (self):
        return '%s%r' % (self.__class__.__name__, self.record)

    @property
    def is_subblock(self):
        return self.record[1]=='SUBBLOCK'
    @property
    def label(self):
        return self.record[2]
    @property
    def data(self):
        return self.record[3]

    def get(self, label):
        if self.label == label:
            if self.is_subblock:
                return self
            return self.data
        if self.is_subblock and self.data is not None:
            l = []
            for entry in self.data:
                r = entry.get(label)
                if isinstance(r, list):
                    l.extend(r)
                elif r is not None:
                    l.append(r)
            if not l:
                return
            return l
    @property
    def offset(self):
        if self._offset is None:
            self._offset = self.ifdentry.value['OffsetScanInformation'][0]
        return self._offset

    def get_size(self):
        """ Return total memory size in bytes needed to fit the entry to memory.
        """
        (entry, type_name, label, data) = self.record
        if type_name=='SUBBLOCK':
            if data is None:
                return 12
            size = 0
            for item in data:
                size += item.get_size()
            return 12 + size
        if type_name=='LONG':
            return 12 + 4
        if type_name=='DOUBLE':
            return 12 + 8
        if type_name == 'ASCII':
            return 12 + len(data) + 1
        raise NotImplementedError (repr(self.record))

    def toarray(self, target = None):
        if target is None:
            target = numpy.zeros((self.get_size(),), dtype=numpy.ubyte)
        dtype = target.dtype
        (entry, type_name, label, data) = self.record
        target[:12] = self.header
        if type_name=='SUBBLOCK':
            if data is not None:
                n = 12
                for item in data:
                    item.toarray(target[n:])
                    n += item.get_size()
        elif type_name == 'ASCII':
            target[12:12+len(data)+1] = numpy.array([data+'\0']).view(dtype=dtype)
        elif type_name == 'LONG':
            target[12:12+4] = numpy.array([data], dtype=numpy.uint32).view(dtype=dtype)
        elif type_name == 'DOUBLE':
            target[12:12+8] = numpy.array([data], dtype=numpy.float64).view(dtype=dtype)
        else:
            raise NotImplementedError (repr(self.record))
        return target

    def tostr(self, tab='', short=False):
        (entry, type_name, label, data) = self.record
        if hasattr(self, 'parent') and self.parent is not None:
            parent_label = self.parent.record[2]
        else:
            parent_label = ''
        if type_name=='SUBBLOCK':
            if data is None:
                return '%s%s %s' % (tab[:-2], label, parent_label)
            l = ['%s%s[size=%s]' % (tab, label, self.get_size ())]
            for item in data:
                if not short or item.data:
                    l.append(item.tostr(tab=tab+'  ', short=short))
            return '\n'.join (l)
        if label.startswith(parent_label):
            label = label[len(parent_label):]
        return '%s%s = %r' % (tab, label, data)

    __str__ = tostr

    def append (self, entry):
        assert self.record[1]=='SUBBLOCK',repr(self.record)
        self.record[3].append(entry)


def scaninfo(ifdentry, debug=True):
    """
    Return LSM scan information.

    Parameters
    ----------
    ifdentry : IFDEntry
    debug: bool
      Enable consistency checks.
    Returns
    -------
    record : ScanInfoEntry
    """
    if ifdentry.tag!=CZ_LSMInfo_tag:
        return
    #assert ifdentry.is_lsm
    n = n1 = ifdentry.value['OffsetScanInformation'][0]
    if not n:
        return
    record = None
    tab = ' '
    while 1:
        entry, type, size = ifdentry.tiff.get_values(n, 'LONG', 3)
        n += 12
        label, type_name = scaninfo_map.get(entry, (None, None))
        if label is None:
            type_name = {0:'SUBBLOCK', 2:'ASCII', 4:'LONG', 5:'DOUBLE'}.get(type)
            if type_name is None:
                raise NotImplementedError(repr((hex (entry), type, size)))
            label = 'ENTRY%s' % (hex(entry))
            #scaninfo_map[entry] = label, type_name
            if debug:
                sys.stderr.write('lsm.scaninfo: undefined %s entry %s in subblock %r\n'\
                                     % (type_name, hex(entry), record.record[2]))
                pass
        if type_name=='SUBBLOCK':
            assert type==0,repr((hex (entry), type, size))
            if label == 'end':
                entry = ScanInfoEntry(entry, type_name, label, None)
                entry.parent = record
                record.append(entry)
                if record.parent is None:
                    break
                record.parent.append(record)
                record = record.parent
            else:
                prev_record = record
                record = ScanInfoEntry(entry, type_name, label, [])
                record.parent = prev_record
            assert size==0,repr((hex (entry), type, size))
            continue
        if type_name=='ASCII':
            assert type==2,repr((hex (entry), type, size))
            value = ifdentry.tiff.get_string (n, size-1)
        elif type_name=='LONG':
            assert type==4,repr((hex (entry), type, size, scaninfo_map[entry]))
            value = ifdentry.tiff.get_long(n)
        elif type_name=='DOUBLE':
            assert type==5,repr((hex (entry), type, size))
            value = ifdentry.tiff.get_double(n)
        else:
            raise NotImplementedError(repr((type_name, hex (entry), type, size)))
        entry = ScanInfoEntry(entry, type_name, label, value)
        entry.parent = record
        n += size
        record.append(entry)
    if debug:
        size = n - n1
        record_size = record.get_size()
        assert size==record_size,repr((size,record_size))
        arr = record.toarray()
        arr2 = ifdentry.tiff.data[n1:n1+size]
        assert (arr==arr2).all()

    record.offset = n1

    return record

class TimeStamps:
    """ Holds LSM time stamps information.
    """
    def __init__(self, ifdentry):
        self.ifdentry = ifdentry
        self._offset = None
        self._header = None
        self._stamps = None

    @property
    def offset(self):
        if self._offset is None:
            self._offset = self.ifdentry.value['OffsetTimeStamps'][0]            
        return self._offset

    @property
    def header(self):
        if self._header is None:
            offset = self.offset
            self._header = self.ifdentry.tiff.get_value(offset, CZ_LSMTimeStamps_header_dtype)            
        return self._header

    @property
    def stamps(self):
        if self._stamps is None:
            n = self.header['NumberTimeStamps']
            self._stamps = self.ifdentry.tiff.get_values(self.offset + self.header.dtype.itemsize, numpy.float64, int(n))
        return self._stamps

    def __str__(self):
        return '%s(stamps=%s)' % (self.__class__.__name__, self.stamps)

    def get_size(self):
        return 8+self.stamps.nbytes

    def toarray(self, target=None):
        sz = self.get_size()
        if target is None:
            target = numpy.zeros((sz,), dtype=numpy.ubyte)
        dtype = target.dtype
        header = numpy.array([sz, self.stamps.size], dtype=numpy.int32).view(dtype=dtype)
        assert header.nbytes==8,repr(header.nbytes)
        data = self.stamps.view(dtype=dtype)
        target[:header.nbytes] = header
        target[header.nbytes:header.nbytes+data.nbytes] = data
        return target

def timestamps(ifdentry, debug=True):
    """
    Return LSM time stamp information.
    """
    if ifdentry.tag!=CZ_LSMInfo_tag:
        return
    #assert ifdentry.is_lsm
    offset = ifdentry.value['OffsetTimeStamps'][0]
    if not offset:
        return None
    r = TimeStamps(ifdentry)
    if debug:
        arr = r.toarray()
        arr2 = ifdentry.tiff.data[offset:offset+arr.nbytes]
        assert (arr==arr2).all()
    return r

class EventEntry:
    
    def __init__ (self, entry_size, time, event_type, unknown, description):
        self.record = entry_size, time, event_type, unknown, description

    @property
    def time (self): return self.record[1]
    @property
    def type (self): return self.record[2]
    @property
    def type_name (self):
        r = {0:'marker', 1:'timer change', 2:'cleach start',
             3:'bleach stop', 4:'trigger'}.get (self.type)
        if r is None:
            r = 'EventType%s' % (self.type)
        return r
    @property
    def description(self): return self.record[4]

    def __str__(self):
        return 'EventEntry(time=%r, type=%r, description=%r)' % (self.time, self.type_name, self.description)

    def get_size(self):
        return 4+8+4+4+len(self.description)+1

    def toarray (self, target = None):
        sz = self.get_size()
        if target is None:
            target = numpy.zeros((sz,), dtype=numpy.ubyte)
        dtype = target.dtype
        target[:4].view(dtype=numpy.uint32)[0] = sz
        target[4:4+8].view(dtype=numpy.float64)[0] = self.time
        target[12:16].view (dtype=numpy.uint32)[0] = self.type
        target[16:20].view (dtype=numpy.uint32)[0] = self.record[3]
        l = len (self.description)
        target[20:20+l] = numpy.array([self.description]).view(dtype=dtype)
        target[20+l] = 0
        return target

class EventList:
    """ Holds LSM event list information.
    """
    def __init__(self, ifdentry):
        self.ifdentry = ifdentry
        self._offset = None
        self._header = None
        self._events = None

    @property
    def offset(self):
        if self._offset is None:
            self._offset = self.ifdentry.value['OffsetEventList'][0]
        return self._offset

    @property
    def header(self):
        if self._header is None:
            self._header = self.ifdentry.tiff.get_value(self.offset, CZ_LSMEventList_header_dtype)
        return self._header

    @property
    def events(self):
        if self._events is None:
            n = self.header['NumberEvents']
            offset = self.offset + self.header.nbytes
            self._events = []
            for i in range(n):
                entry_size = self.ifdentry.tiff.get_value (offset, numpy.uint32)
                time = self.ifdentry.tiff.get_value(offset+4, numpy.float64)
                event_type = self.ifdentry.tiff.get_value(offset + 4+8, numpy.uint32)
                unknown = self.ifdentry.tiff.get_value(offset + 4+8+4, numpy.uint32)
                descr = self.ifdentry.tiff.get_string(offset + 4+8+4+4)
                self._events.append(EventEntry(entry_size, time, event_type, unknown, descr))
                offset += entry_size
        return self._events

    def __str__ (self):
        return '%s(events=[%s])' % (self.__class__.__name__, ','.join(map(str,self.events)))

    def get_size(self):
        s = self.header.nbytes
        for event in self.events:
            s += event.get_size()
        return s

    def toarray (self, target=None):
        sz = self.get_size()
        if target is None:
            target = numpy.zeros((sz,), dtype=numpy.ubyte)
        dtype = target.dtype     
        header = numpy.array([sz, len (self.events)], dtype=numpy.int32).view(dtype=dtype)
        target[:header.nbytes] = header
        offset = header.nbytes
        for event in self.events:
            event.toarray(target[offset:])
            offset += event.get_size()
        return target

def eventlist(ifdentry, debug=True):
    """
    Return LSM event list information.
    """
    if ifdentry.tag!=CZ_LSMInfo_tag:
        return
    #assert ifdentry.is_lsm
    offset = ifdentry.value['OffsetEventList'][0]
    if not offset:
        return None
    r = EventList(ifdentry)
    if debug:
        arr = r.toarray()
        arr2 = ifdentry.tiff.data[offset:offset+arr.nbytes]
        assert (arr==arr2).all()
    return r

class ChannelWavelength:
    """ Holds LSM channel wavelength information.
    """
    def __init__ (self, ifdentry):
        self.ifdentry = ifdentry
        self._offset = None
        self._ranges = None

    @property
    def offset(self):
        if self._offset is None:
            self._offset = self.ifdentry.value['OffsetChannelWavelength'][0]            
        return self._offset

    @property
    def ranges (self):
        if self._ranges is None:
            self._ranges = []
            offset = self.offset
            n = self.ifdentry.tiff.get_value (offset, numpy.int32)
            for i in range(n):
                start, end = self.ifdentry.tiff.get_values(offset + 4 + i*(8+8), numpy.float64, 2)
                self._ranges.append((start, end))
        return self._ranges

    def __str__ (self):
        return '%s (ranges=%s)' % (self.__class__.__name__, self.ranges)

    def get_size(self):
        return 4 + len(self.ranges)*16

    def toarray (self, target=None):
        sz = self.get_size()
        if target is None:
            target = numpy.zeros((sz,), dtype=numpy.ubyte)
        dtype = target.dtype     
        target[:4].view(dtype=numpy.int32)[0] = len(self.ranges)
        data = numpy.array(self.ranges).ravel()
        target[4:4+data.nbytes] = data.view (dtype=dtype)
        return target

def channelwavelength(ifdentry, debug=True):
    """
    Return LSM wavelength range information.
    """
    if ifdentry.tag!=CZ_LSMInfo_tag:
        return
    #assert ifdentry.is_lsm
    offset = ifdentry.value['OffsetChannelWavelength'][0]
    if not offset:
        return None
    r = ChannelWavelength(ifdentry)
    if debug:
        arr = r.toarray()
        arr2 = ifdentry.tiff.data[offset:offset+arr.nbytes]
        assert (arr==arr2).all()
    return r

class ChannelColors:
    """ Holds LSM channel name and color information.
    """
    def __init__ (self, ifdentry):
        self.ifdentry = ifdentry
        self._offset = None
        self._header = None
        self._names = None
        self._colors = None
        self._mono = None

    @property
    def offset(self):
        if self._offset is None:
            self._offset = self.ifdentry.value['OffsetChannelColors'][0]            
        return self._offset

    @property
    def mono (self):
        if self._mono is None:
            self._mono = not not self.header[5]
        return self._mono

    @property
    def header (self):
        if self._header is None:
            self._header = self.ifdentry.tiff.get_values(self.offset, numpy.int32, 10)
            sz = self._header[0]
        return self._header

    @property
    def names(self):
        if self._names is None:
            header = self.header
            n = header[2]
            offset = self.offset + header[4] + 4
            self._names = []
            for i in range(n):
                name = self.ifdentry.tiff.get_string(offset)
                offset += len(name) + 1 + 4
                self._names.append(name)
        return self._names
    
    @property
    def colors(self):
        if self._colors is None:
            header = self.header
            n = header[1]
            offset = self.offset + header[3]
            self._colors = []
            for i in range(n):
                color = self.ifdentry.tiff.get_values(offset, numpy.uint8, 4)
                offset += color.nbytes
                self._colors.append(tuple(color))
        return self._colors

    def __str__(self):
        return '%s (names=%s, colors=%s)' % (self.__class__.__name__, self.names, self.colors)

    def get_size(self):
        s = 10*4
        for name in self.names:
            s += len(name) + 1 + 4
        for color in self.colors:
            s += len(color)
        return s

    def toarray (self, target=None):
        sz = self.get_size()
        if target is None:
            target = numpy.zeros((sz,), dtype=numpy.ubyte)
        dtype = target.dtype
        header = numpy.array([sz,len (self.colors), len (self.names), 0, 0, self.mono, 0,0,0,0], dtype=numpy.int32)
        names = ''
        for name in self.names:
            names += '\x04\0\0\0' + name + '\x00'
        noffset = sz - len(names)
        names = numpy.array([names]).view (dtype=dtype)
        colors = numpy.array (self.colors, dtype=numpy.uint8).ravel()
        coffset = noffset - colors.nbytes
        header[3] = coffset
        header[4] = noffset

        target[:header.nbytes] = header.view(dtype=dtype)
        target[coffset:coffset + colors.nbytes] = colors
        target[noffset:noffset + names.nbytes] = names
        return target

def channelcolors(ifdentry, debug=True):
    """
    Return LSM channel name and color information.
    """
    if ifdentry.tag!=CZ_LSMInfo_tag:
        return
    #assert ifdentry.is_lsm
    offset = ifdentry.value['OffsetChannelColors'][0]
    if not offset:
        return None
    r = ChannelColors(ifdentry)
    if debug:
        arr = r.toarray()
        arr2 = ifdentry.tiff.data[offset:offset+arr.nbytes]
        assert (arr==arr2).all()
    return r

class ChannelFactors:

    def __init__(self, ifdentry):
        self.ifdentry = ifdentry
        self._offset = None
        self._data = None

    @property
    def offset(self):
        if self._offset is None:
            self._offset = self.ifdentry.value['OffsetChannelFactors'][0]            
        return self._offset

    @property
    def data(self):
        if self._data is None:
            n = self.ifdentry.tiff.get_value(self.offset, numpy.int32)
            sz = 4 + n # n should be equal to 32*nofchannels
            self._data = self.ifdentry.tiff.get_values(self.offset, numpy.ubyte, sz)
        return self._data

    def __str__(self):
        return '%s(size=%r, offset=%r)' % (self.__class__.__name__, self.get_size(), self.offset)

    def get_size(self):
        return self.data.nbytes

    def toarray(self, target=None):
        sz = self.get_size()
        if target is None:
            target = numpy.zeros((sz,), dtype=numpy.ubyte)
        dtype = target.dtype
        target[:self.data.nbytes] = self.data.view(dtype=dtype)
        return target

def channelfactors(ifdentry, debug=True):
    if ifdentry.tag!=CZ_LSMInfo_tag:
        return
    #assert ifdentry.is_lsm
    offset = ifdentry.value['OffsetChannelFactors'][0]
    if not offset:
        return
    r = ChannelFactors(ifdentry)
    if debug:
        arr = r.toarray()
        arr2 = ifdentry.tiff.data[offset:offset+arr.nbytes]
        assert (arr==arr2).all()        
    return r

class ChannelDataTypes:

    def __init__(self, ifdentry):
        self.ifdentry = ifdentry
        self._offset = None
        self._data = None

    @property
    def offset(self):
        if self._offset is None:
            self._offset = self.ifdentry.value['OffsetChannelDataTypes'][0]            
        return self._offset

    @property
    def data(self):
        if self._data is None:
            channels = self.ifdentry.value['DimensionChannels']
            sz = channels * 4
            self._data = self.ifdentry.tiff.get_values(self.offset, numpy.ubyte, sz)
        return self._data

    def __str__(self):
        return '%s(size=%r, offset=%r)' % (self.__class__.__name__, self.get_size(), self.offset)

    def get_size(self):
        return self.data.nbytes

    def toarray(self, target=None):
        sz = self.get_size()
        if target is None:
            target = numpy.zeros((sz,), dtype=numpy.ubyte)
        dtype = target.dtype
        target[:self.data.nbytes] = self.data.view(dtype=dtype)
        return target

def channeldatatypes(ifdentry, debug=True):
    if ifdentry.tag!=CZ_LSMInfo_tag:
        return
    #assert ifdentry.is_lsm
    offset = ifdentry.value['OffsetChannelDataTypes'][0]
    if not offset:
        return
    r = ChannelDataTypes(ifdentry)
    if debug:
        arr = r.toarray()
        arr2 = ifdentry.tiff.data[offset:offset+arr.nbytes]
        assert (arr==arr2).all()        
    return r

class OffsetData:

    def __init__(self, ifdentry, offset_name):
        self.ifdentry = ifdentry
        self.offset_name = offset_name
        self._offset = None
        self._data = None

    @property
    def offset(self):
        if self._offset is None:
            self._offset = self.ifdentry.value[self.offset_name][0]            
        return self._offset

    @property
    def data(self):
        if self._data is None:
            sz = self.ifdentry.tiff.get_value(self.offset, numpy.uint32)
            self._data = self.ifdentry.tiff.get_values(self.offset, numpy.ubyte, sz)
        return self._data

    def __str__(self):
        return '%s(name=%r, size=%r, offset=%r)' % (self.__class__.__name__, self.offset_name, self.get_size(), self.offset)

    def get_size(self):
        return self.data.nbytes

    def toarray(self, target=None):
        sz = self.get_size()
        if target is None:
            target = numpy.zeros((sz,), dtype=numpy.ubyte)
        dtype = target.dtype
        target[:self.data.nbytes] = self.data.view(dtype=dtype)
        return target
        

def offsetdata(ifdentry, offset_name, debug=True):
    if ifdentry.tag!=CZ_LSMInfo_tag:
        return
    #assert ifdentry.is_lsm
    offset = ifdentry.value[offset_name][0]
    if not offset:
        return
    r = OffsetData(ifdentry, offset_name)
    if debug:
        arr = r.toarray()
        arr2 = ifdentry.tiff.data[offset:offset+arr.nbytes]
        assert (arr==arr2).all()        
    return r

class DrawingElement:

    def __init__(self, ifdentry, offset_name):
        self.ifdentry = ifdentry
        self.offset_name = offset_name
        self._offset = None
        self._data = None

    @property
    def offset(self):
        if self._offset is None:
            self._offset = self.ifdentry.value[self.offset_name][0]            
        return self._offset

    @property
    def data(self):
        if self._data is None:
            n = self.ifdentry.tiff.get_value(self.offset, numpy.int32)
            sz = self.ifdentry.tiff.get_value(self.offset+4, numpy.int32)
            self._data = self.ifdentry.tiff.get_values(self.offset, numpy.ubyte, sz)
        return self._data

    def __str__(self):
        return '%s(name=%r, size=%r, offset=%r)' % (self.__class__.__name__, self.offset_name, self.get_size(), self.offset)

    def get_size(self):
        return self.data.nbytes

    def toarray(self, target=None):
        sz = self.get_size()
        if target is None:
            target = numpy.zeros((sz,), dtype=numpy.ubyte)
        dtype = target.dtype
        target[:self.data.nbytes] = self.data.view(dtype=dtype)
        return target

def drawingelement(ifdentry, offset_name, debug=True):
    if ifdentry.tag!=CZ_LSMInfo_tag:
        return
    #assert ifdentry.is_lsm
    offset = ifdentry.value[offset_name][0]
    if not offset:
        return
    r = DrawingElement(ifdentry, offset_name)
    if debug:
        arr = r.toarray()
        arr2 = ifdentry.tiff.data[offset:offset+arr.nbytes]
        assert (arr==arr2).all()        
    return r

class LookupTable:

    def __init__(self, ifdentry, offset_name):
        self.ifdentry = ifdentry
        self.offset_name = offset_name
        self._offset = None
        self._data = None

    @property
    def offset(self):
        if self._offset is None:
            self._offset = self.ifdentry.value[self.offset_name][0]            
        return self._offset

    @property
    def data(self):
        if self._data is None:
            sz = self.ifdentry.tiff.get_value(self.offset, numpy.uint32)
            self._data = self.ifdentry.tiff.get_values(self.offset, numpy.ubyte, sz)
        return self._data

    def __str__(self):
        nsubblocks = self.ifdentry.tiff.get_value(self.offset+4, numpy.uint32)
        nchannels = self.ifdentry.tiff.get_value(self.offset+8, numpy.uint32)
        return '%s(name=%r, size=%r, subblocks=%r, channels=%r, offset=%r)' \
               % (self.__class__.__name__, self.offset_name, self.get_size(),
                  nsubblocks, nchannels, self.offset)

    def get_size(self):
        return self.data.nbytes

    def toarray(self, target=None):
        sz = self.get_size()
        if target is None:
            target = numpy.zeros((sz,), dtype=numpy.ubyte)
        dtype = target.dtype
        target[:self.data.nbytes] = self.data.view(dtype=dtype)
        return target
        

def lookuptable(ifdentry, offset_name, debug=True):
    if ifdentry.tag!=CZ_LSMInfo_tag:
        return
    offset = ifdentry.value[offset_name][0]
    if not offset:
        return
    r = LookupTable(ifdentry, offset_name)
    if debug:
        arr = r.toarray()
        arr2 = ifdentry.tiff.data[offset:offset+arr.nbytes]
        assert (arr==arr2).all()        
    return r

inputlut = lambda ifdentry, debug=True: lookuptable(ifdentry, 'OffsetInputLut', debug=debug)
outputlut = lambda ifdentry, debug=True: lookuptable(ifdentry, 'OffsetOutputLut', debug=debug)
vectoroverlay = lambda ifdentry, debug=True: drawingelement(ifdentry, 'OffsetVectorOverlay', debug=debug)
roi = lambda ifdentry, debug=True: drawingelement(ifdentry, 'OffsetRoi', debug=debug)
bleachroi = lambda ifdentry, debug=True: drawingelement(ifdentry, 'OffsetBleachRoi', debug=debug)
meanofroisoverlay = lambda ifdentry, debug=True: drawingelement(ifdentry, 'OffsetMeanOfRoisOverlay', debug=debug)
topoisolineoverlay = lambda ifdentry, debug=True: drawingelement(ifdentry, 'OffsetTopoIsolineOverlay', debug=debug)
topoprofileoverlay = lambda ifdentry, debug=True: drawingelement(ifdentry, 'OffsetTopoProfileOverlay', debug=debug)
linescanoverlay = lambda ifdentry, debug=True: drawingelement(ifdentry, 'OffsetLinescanOverlay', debug=debug)

CZ_LSMOffsetField_readers = dict(
    OffsetChannelWavelength = channelwavelength,
    OffsetTimeStamps=timestamps,
    OffsetEventList=eventlist,
    OffsetChannelColors = channelcolors,
    OffsetScanInformation = scaninfo,
    OffsetInputLut = inputlut,
    OffsetOutputLut = outputlut,
    OffsetChannelFactors = channelfactors,
    OffsetChannelDataTypes = channeldatatypes,
    OffsetUnmixParameters = lambda ifdentry, debug=True: offsetdata(ifdentry, 'OffsetUnmixParameters', debug=debug),
    OffsetNextRecording = lambda ifdentry, debug=True: offsetdata(ifdentry, 'OffsetNextRecording', debug=debug),
    OffsetKsData = lambda ifdentry, debug=True: offsetdata(ifdentry, 'OffsetKsData', debug=debug),

    OffsetVectorOverlay = vectoroverlay,
    OffsetRoi = roi,
    OffsetBleachRoi = bleachroi,
    OffsetMeanOfRoisOverlay = meanofroisoverlay,
    OffsetTopoIsolineOverlay = topoisolineoverlay,
    OffsetTopoProfileOverlay = topoprofileoverlay,
    OffsetLinescanOverlay = linescanoverlay,
    )

CZ_LSMInfo_dtype_fields = [
    ('MagicNumber',numpy.uint32),
    ('StructureSize',numpy.int32),
    ('DimensionX',numpy.int32),
    ('DimensionY',numpy.int32),
    ('DimensionZ',numpy.int32),
    ('DimensionChannels', numpy.int32),
    ('DimensionTime', numpy.int32),
    ('SDataType', numpy.int32), # 1: uint8, 2: uint12, 5: float32, 0: see OffsetChannelDataTypes
    ('ThumbnailX', numpy.int32),
    ('ThumbnailY', numpy.int32),
    ('VoxelSizeX', numpy.float64),
    ('VoxelSizeY', numpy.float64),
    ('VoxelSizeZ', numpy.float64),
    ('OriginX', numpy.float64),
    ('OriginY', numpy.float64),
    ('OriginZ', numpy.float64),
    ('ScanType', numpy.uint16),
    # 0:xyz-scan, 1:z-scan, 2:line-scan 3:time xy,
    # 4: time xz (ver>=2.0), 5: time mean roi (ver>=2.0),
    # 6: time xyz (ver>=2.3), 7: spline scan (ver>=2.5),
    # 8: spline plane xz (ver>=2.5)
    # 9: time spline plane xz (ver>=2.5)
    # 10: point mode (ver>=3.0)
    ('SpectralScan', numpy.uint16), # 0:off, 1:on (ver>=3.0)
    ('DataType',numpy.uint32), # 0: original, 1: calculated, 2: animation
    ('OffsetVectorOverlay', numpy.uint32),
    ('OffsetInputLut',numpy.uint32),
    ('OffsetOutputLut',numpy.uint32),
    ('OffsetChannelColors',numpy.uint32),
    ('TimeInterval',numpy.float64),
    ('OffsetChannelDataTypes',numpy.uint32),
    ('OffsetScanInformation',numpy.uint32),
    ('OffsetKsData',numpy.uint32),
    ('OffsetTimeStamps',numpy.uint32),
    ('OffsetEventList',numpy.uint32),
    ('OffsetRoi',numpy.uint32),
    ('OffsetBleachRoi',numpy.uint32),
    ('OffsetNextRecording',numpy.uint32),
    ('DisplayAspectX',numpy.float64),
    ('DisplayAspectY',numpy.float64),
    ('DisplayAspectZ',numpy.float64),
    ('DisplayAspectTime',numpy.float64),
    ('OffsetMeanOfRoisOverlay',numpy.uint32),
    ('OffsetTopoIsolineOverlay', numpy.uint32),
    ('OffsetTopoProfileOverlay', numpy.uint32),
    ('OffsetLinescanOverlay', numpy.uint32),
    ('ToolbarFlags', numpy.uint32),
    ('OffsetChannelWavelength', numpy.uint32),
    ('OffsetChannelFactors', numpy.uint32),
    ('ObjectiveSphereCorrection',numpy.float64),
    ('OffsetUnmixParameters',numpy.uint32),
    #('Reserved', numpy.dtype('(69,)u4')), # depends on the version of LSM file
    ]

CZ_LSMInfo_dtype_fields_size = 0
for item in CZ_LSMInfo_dtype_fields:
    CZ_LSMInfo_dtype_fields_size += item[1]().itemsize

CZ_LSMTimeStamps_header_dtype = numpy.dtype([
        ('Size', numpy.int32),
        ('NumberTimeStamps', numpy.int32),
        # ('TimeStamp<N>', numpy.float64)
        ])

CZ_LSMEventList_header_dtype = numpy.dtype ([
        ('Size', numpy.int32),
        ('NumberEvents', numpy.int32),
        # ('Event<N>', EventListEntry)
        ])

scaninfo_map = {
    0x0ffffffff: ( 'end', 'SUBBLOCK'),
    0x010000000: ( 'recording','SUBBLOCK'),
    0x010000001: ( 'recording name','ASCII'),
    0x010000002: ( 'recording description','ASCII'),
    0x010000003: ( 'recording notes','ASCII'),
    0x010000004: ( 'recording objective','ASCII'),
    0x010000005: ( 'recording processing summary','ASCII'),
    0x010000006: ( 'recording special scan mode','ASCII'),
    0x010000007: ( 'recording scan type','ASCII'),
    0x010000008: ( 'recording scan mode','ASCII'),
    0x010000009: ( 'recording number of stacks','LONG'),
    0x01000000a: ( 'recording lines per plane','LONG'),
    0x01000000b: ( 'recording samples per line','LONG'),
    0x01000000c: ( 'recording planes per volume','LONG'),
    0x01000000d: ( 'recording images width','LONG'),
    0x01000000e: ( 'recording images height','LONG'),
    0x01000000f: ( 'recording images number planes','LONG'),
    0x010000010: ( 'recording images number stacks','LONG'),
    0x010000011: ( 'recording images number channels','LONG'),
    0x010000012: ( 'recording linescan xy size','LONG'),
    0x010000013: ( 'recording scan direction','LONG'),
    0x010000014: ( 'recording time series','LONG'),
    0x010000015: ( 'recording original scan data','LONG'),
    0x010000016: ( 'recording zoom x','DOUBLE'),
    0x010000017: ( 'recording zoom y','DOUBLE'),
    0x010000018: ( 'recording zoom z','DOUBLE'),
    0x010000019: ( 'recording sample 0x','DOUBLE'),
    0x01000001a: ( 'recording sample 0y','DOUBLE'),
    0x01000001b: ( 'recording sample 0z','DOUBLE'),
    0x01000001c: ( 'recording sample spacing','DOUBLE'),
    0x01000001d: ( 'recording line spacing','DOUBLE'),
    0x01000001e: ( 'recording plane spacing','DOUBLE'),
    0x01000001f: ( 'recording plane width','DOUBLE'),
    0x010000020: ( 'recording plane height','DOUBLE'),
    0x010000021: ( 'recording volume depth','DOUBLE'),
    #0x010000022: ( 'recording ',''),
    0x010000023: ( 'recording nutation','DOUBLE'),
    #0x010000024: ( 'recording ',''),
    #0x010000025: ( 'recording ',''),
    #0x010000026: ( 'recording ',''),
    #0x010000027: ( 'recording ',''),
    #0x010000028: ( 'recording ',''),
    #0x010000029: ( 'recording ',''),
    #0x01000002a: ( 'recording ',''),
    #0x01000002b: ( 'recording ',''),
    #0x01000002c: ( 'recording ',''),
    #0x01000002d: ( 'recording ',''),
    #0x01000002e: ( 'recording ',''),
    #0x01000002f: ( 'recording ',''),
    #0x010000030: ( 'recording ',''),
    #0x010000031: ( 'recording ',''),
    #0x010000032: ( 'recording ',''),
    #0x010000033: ( 'recording ',''),
    0x010000034: ( 'recording rotation','DOUBLE'),
    0x010000035: ( 'recording precession','DOUBLE'),
    0x010000036: ( 'recording sample 0time','DOUBLE'),
    0x010000037: ( 'recording start scan trigger in','ASCII'),
    0x010000038: ( 'recording start scan trigger out','ASCII'),
    0x010000039: ( 'recording start scan event','LONG'),
    #0x01000003a: ( 'recording ',''),
    #0x01000003b: ( 'recording ',''),
    #0x01000003c: ( 'recording ',''),
    #0x01000003d: ( 'recording ',''),
    #0x01000003e: ( 'recording ',''),
    #0x01000003f: ( 'recording ',''),
    0x010000040: ( 'recording start scan time','DOUBLE'),
    0x010000041: ( 'recording stop scan trigger in','ASCII'),
    0x010000042: ( 'recording stop scan trigger out','ASCII'),
    0x010000043: ( 'recording stop scan event','LONG'),
    0x010000044: ( 'recording start scan time','DOUBLE'),
    0x010000045: ( 'recording use rois','LONG'),
    0x010000046: ( 'recording use reduced memory rois','LONG'),
    0x010000047: ( 'recording user','ASCII'),
    0x010000048: ( 'recording usebccorrection','LONG'),
    0x010000049: ( 'recording positionbccorrection1','DOUBLE'),
    #0x01000004a: ( 'recording ',''),
    #0x01000004b: ( 'recording ',''),
    #0x01000004c: ( 'recording ',''),
    #0x01000004d: ( 'recording ',''),
    #0x01000004e: ( 'recording ',''),
    #0x01000004f: ( 'recording ',''),
    0x010000050: ( 'recording positionbccorrection2','DOUBLE'),
    0x010000051: ( 'recording interpolationy','LONG'),
    0x010000052: ( 'recording camera binning','LONG'),
    0x010000053: ( 'recording camera supersampling','LONG'),
    0x010000054: ( 'recording camera frame width','LONG'),
    0x010000055: ( 'recording camera frame height','LONG'),
    0x010000056: ( 'recording camera offsetx','DOUBLE'),
    0x010000057: ( 'recording camera offsety','DOUBLE'),
    #0x010000058: ( 'recording ',''),
    0x010000059: ( 'recording rt binning','LONG'),
    0x01000005a: ( 'recording rt frame width','LONG'),
    0x01000005b: ( 'recording rt frame height','LONG'),
    0x01000005c: ( 'recording rt region width','LONG'),
    0x01000005d: ( 'recording rt region height','LONG'),
    0x01000005e: ( 'recording rt offsetx','DOUBLE'),
    0x01000005f: ( 'recording rt offsety','DOUBLE'),
    0x010000060: ( 'recording rt zoom','DOUBLE'),
    0x010000061: ( 'recording rt lineperiod','DOUBLE'),
    0x010000062: ( 'recording prescan','LONG'),
    0x010000063: ( 'recording scan directionz','LONG'),
    #0x010000064: ( 'recording ','LONG'),
    0x030000000: ( 'lasers','SUBBLOCK'), 
    0x050000000: ( 'laser','SUBBLOCK'),
    0x050000001: ( 'laser name','ASCII'),
    0x050000002: ( 'laser acquire','LONG'),
    0x050000003: ( 'laser power','DOUBLE'),
    0x020000000: ( 'tracks','SUBBLOCK'), 
    0x040000000: ( 'track','SUBBLOCK'),
    0x040000001: ( 'track multiplex type','LONG'),
    0x040000002: ( 'track multiplex order','LONG'),
    0x040000003: ( 'track sampling mode','LONG'),
    0x040000004: ( 'track sampling method','LONG'),
    0x040000005: ( 'track sampling number','LONG'),
    0x040000006: ( 'track acquire','LONG'),
    0x040000007: ( 'track sample observation time','DOUBLE'),
    #
    0x04000000b: ( 'track time between stacks','DOUBLE'),
    0x04000000c: ( 'track name','ASCII'),
    0x04000000d: ( 'track collimator1 name','ASCII'),
    0x04000000e: ( 'track collimator1 position','LONG'),
    0x04000000f: ( 'track collimator2 name','ASCII'),
    0x040000010: ( 'track collimator2 position','LONG'),
    0x040000011: ( 'track is bleach track','LONG'),
    0x040000012: ( 'track is bleach after scan number','LONG'),
    0x040000013: ( 'track bleach scan number','LONG'),
    0x040000014: ( 'track trigger in','ASCII'),
    0x040000015: ( 'track trigger out','ASCII'),
    0x040000016: ( 'track is ratio track','LONG'),
    0x040000017: ( 'track bleach count','LONG'),
    0x040000018: ( 'track spi center wavelength','DOUBLE'),
    0x040000019: ( 'track pixel time','DOUBLE'),
    0x040000020: ( 'track id condensor frontlens','ASCII'),
    0x040000021: ( 'track condensor frontlens','LONG'),
    0x040000022: ( 'track id field stop','ASCII'),
    0x040000023: ( 'track field stop value','DOUBLE'),
    0x040000024: ( 'track id condensor aperture','ASCII'),
    0x040000025: ( 'track condensor aperture','DOUBLE'),
    0x040000026: ( 'track id condensor revolver','ASCII'),
    0x040000027: ( 'track condensor filter','ASCII'),
    0x040000028: ( 'track id transmission filter1','ASCII'),
    0x040000029: ( 'track id transmission1','DOUBLE'),
    0x040000030: ( 'track id transmission filter2','ASCII'),
    0x040000031: ( 'track if transmission2','DOUBLE'),
    0x040000032: ( 'track repeat bleach','LONG'),
    0x040000033: ( 'track enable spot bleach pos','LONG'),
    0x040000034: ( 'track spot bleach posx','DOUBLE'),
    0x040000035: ( 'track spot bleach posy','DOUBLE'),
    0x040000036: ( 'track bleach position z','DOUBLE'),
    0x040000037: ( 'track id tubelens','ASCII'),
    0x040000038: ( 'track id tubelens position','ASCII'),
    0x040000039: ( 'track transmitted light','DOUBLE'),
    0x04000003a: ( 'track reflected light','DOUBLE'),
    0x04000003b: ( 'track simultan grab and bleach','LONG'),
    0x04000003c: ( 'track bleach pixel time','DOUBLE'),

    0x060000000: ( 'detection channels','SUBBLOCK'),
    0x070000000: ( 'detection channel','SUBBLOCK'),
    0x070000001: ( 'detection channel integration mode','LONG'),
    0x070000002: ( 'detection channel special mode','LONG'),
    0x070000003: ( 'detection channel detector gain first','DOUBLE'),
    0x070000004: ( 'detection channel detector gain last','DOUBLE'),
    0x070000005: ( 'detection channel amplifier gain first','DOUBLE'),    
    0x070000006: ( 'detection channel amplifier gain last','DOUBLE'),
    0x070000007: ( 'detection channel amplifier offs first','DOUBLE'),
    0x070000008: ( 'detection channel amplifier offs last','DOUBLE'),
    0x070000009: ( 'detection channel pinhole diameter','DOUBLE'),
    0x07000000a: ( 'detection channel counting trigger','DOUBLE'),
    0x07000000b: ( 'detection channel acquire','LONG'),
    0x07000000c: ( 'detection channel detector name','ASCII'),
    0x07000000d: ( 'detection channel amplifier name','ASCII'),
    0x07000000e: ( 'detection channel pinhole name','ASCII'),
    0x07000000f: ( 'detection channel filter set name','ASCII'),
    0x070000010: ( 'detection channel filter name','ASCII'),
    0x070000013: ( 'detection channel integrator name','ASCII'),
    0x070000014: ( 'detection channel detection channel name','ASCII'),
    0x070000015: ( 'detection channel detector gain bc1','DOUBLE'),
    0x070000016: ( 'detection channel detector gain bc2','DOUBLE'),
    0x070000017: ( 'detection channel amplifier gain bc1','DOUBLE'),
    0x070000018: ( 'detection channel amplifier gain bc2','DOUBLE'),
    0x070000019: ( 'detection channel amplifier offs bc1','DOUBLE'),
    0x070000020: ( 'detection channel amplifier offs bc2','DOUBLE'),
    0x070000021: ( 'detection channel spectral scan channels','LONG'),
    0x070000022: ( 'detection channel spi wavelength start','DOUBLE'),
    0x070000023: ( 'detection channel spi wavelength end','DOUBLE'),
    0x070000026: ( 'detection channel dye name','ASCII'),
    0x070000027: ( 'detection channel dye folder','ASCII'),

    0x080000000: ( 'illumination channels','SUBBLOCK'),
    0x090000000: ( 'illumination channel','SUBBLOCK'),
    0x090000001: ( 'illumination channel name','ASCII'),
    0x090000002: ( 'illumination channel power','DOUBLE'),
    0x090000003: ( 'illumination channel wavelength','DOUBLE'),
    0x090000004: ( 'illumination channel aquire','LONG'),
    0x090000005: ( 'illumination channel detection channel name','ASCII'),
    0x090000006: ( 'illumination channel power bc1','DOUBLE'),
    0x090000007: ( 'illumination channel power bc2','DOUBLE'),

    0x0A0000000: ( 'beam splitters','SUBBLOCK'),
    0x0B0000000: ( 'beam splitter','SUBBLOCK'),
    0x0B0000001: ( 'beam splitter filter set','ASCII'),
    0x0B0000002: ( 'beam splitter filter','ASCII'),
    0x0B0000003: ( 'beam splitter name','ASCII'),

    0x0C0000000: ( 'data channels','SUBBLOCK'),
    0x0D0000000: ( 'data channel','SUBBLOCK'),
    0x0D0000001: ( 'data channel name','ASCII'),
    #0x0D0000002: ( 'data channel',''),
    0x0D0000003: ( 'data channel acquire','LONG'),
    0x0D0000004: ( 'data channel color','LONG'),
    0x0D0000005: ( 'data channel sampletype','LONG'),
    0x0D0000006: ( 'data channel bitspersample','LONG'),
    0x0D0000007: ( 'data channel ratio type','LONG'),
    0x0D0000008: ( 'data channel ratio track1','LONG'),
    0x0D0000009: ( 'data channel ratio track2','LONG'),
    0x0D000000a: ( 'data channel ratio channel1','ASCII'),
    0x0D000000b: ( 'data channel ratio channel2','ASCII'),
    0x0D000000c: ( 'data channel ratio const1','DOUBLE'),
    0x0D000000d: ( 'data channel ratio const2','DOUBLE'),
    0x0D000000e: ( 'data channel ratio const3','DOUBLE'),
    0x0D000000f: ( 'data channel ratio const4','DOUBLE'),
    0x0D0000010: ( 'data channel ratio const5','DOUBLE'),
    0x0D0000011: ( 'data channel ratio const6','DOUBLE'),
    0x0D0000012: ( 'data channel ratio first images1','LONG'),
    0x0D0000013: ( 'data channel ratio first images2','LONG'),
    0x0D0000014: ( 'data channel dye name','ASCII'),
    0x0D0000015: ( 'data channel dye folder','ASCII'),
    0x0D0000016: ( 'data channel spectrum','ASCII'),
    0x0D0000017: ( 'data channel acquire','LONG'),

    0x011000000: ( 'timers','SUBBLOCK'),
    0x012000000: ( 'timer','SUBBLOCK'), 
    0x012000001: ( 'timer name','ASCII'), 
    0x012000003: ( 'timer interval','DOUBLE'), 
    0x012000004: ( 'timer trigger in','ASCII'), 
    0x012000005: ( 'timer trigger out','ASCII'),
    0x012000006: ( 'timer activation time','DOUBLE'),
    0x012000007: ( 'timer activation number','LONG'), 

    0x013000000: ( 'markers','SUBBLOCK'), 
    0x014000000: ( 'marker','SUBBLOCK'),
    0x014000001: ( 'marker name','ASCII'),
    0x014000002: ( 'marker description','ASCII'),
    0x014000003: ( 'marker trigger in','ASCII'),
    0x014000004: ( 'marker trigger out','ASCII'),

}




"""
Defines data for TIFF manipulations.

"""

__all__ = ['type2name', 'name2type', 'type2bytes', 'type2dtype',
           'tag_value2name', 'tag_name2value', 'tag_value2type',
           'LittleEndianNumpyDTypes', 'BigEndianNumpyDTypes',
           'default_tag_values', 'sample_format_map']

import numpy

#<TagName> <Hex> <Type> <Number of values>
tag_info = '''
# standard tags:
NewSubfileType FE LONG 1
SubfileType FF SHORT 1
ImageWidth 100 SHORT|LONG 1
ImageLength 101 SHORT|LONG 1
BitsPerSample 102 SHORT SamplesPerPixel
Compression 103 SHORT 1
  Uncompressed 1
  CCITT1D 2
  Group3Fax 3
  Group4Fax 4
  LZW 5
  JPEG 6
  PackBits 32773
PhotometricInterpretation 106 SHORT 1
  WhiteIsZero 0
  BlackIsZero 1
  RGB 2
  RGBPalette 3
  TransparencyMask 4
  CMYK 5
  YCbCr 6
  CIELab 8
Threshholding 107 SHORT 1
CellWidth 108 SHORT 1
CellLength 109 SHORT 1
FillOrder 10A SHORT 1
DocumentName 10D ASCII
ImageDescription 10E ASCII
Make 10F ASCII
Model 110 ASCII
StripOffsets 111 SHORT|LONG StripsPerImage
Orientation 112 SHORT 1
  TopLeft 1
  TopRight 2
  BottomRight 3
  BottomLeft 4
  LeftTop 5
  RightTop 6
  RightBottom 7
  LeftBottom 8
SamplesPerPixel 115 SHORT 1
RowsPerStrip 116 SHORT|LONG 1
StripByteCounts 117 LONG|SHORT StripsPerImage
MinSampleValue 118 SHORT SamplesPerPixel
MaxSampleValue 119 SHORT SamplesPerPixel
XResolution 11A RATIONAL 1
YResolution 11B RATIONAL 1
PlanarConfiguration 11C SHORT 1
  Chunky 1
  Planar 2
PageName 11D ASCII
XPosition 11E DOUBLE
YPosition 11F DOUBLE
FreeOffsets 120 LONG
FreeByteCounts 121 LONG
GrayResponseUnit 122 SHORT 1
GrayResponseCurve 123 SHORT 2**BitsPerSample
T4Options 124 LONG 1
T6Options 125 LONG 1
ResolutionUnit 128 SHORT 1
PageNumber 129 SHORT 2
TransferFunction 12D SHORT (1|SamplesPerPixel)*2**BitsPerSample
Software 131 ASCII
DateTime 132 ASCII 20
Artist 13B ASCII
HostComputer 13C ASCII
Predictor 13D SHORT 1
WhitePoint 13E RATIONAL 2
PrimaryChromaticities 13F RATIONAL 6
ColorMap 140 SHORT 3*(2**BitsPerSample)
HalftoneHints 141 SHORT 2
TileWidth 142 SHORT|LONG 1
TileLength 143 SHORT|LONG 1
TileOffsets 144 LONG TilesPerImage
TileByteCounts 145 SHORT|LONG TilesPerImage
InkSet 14C SHORT 1
InkNames 14D ASCII <total number of chars in all ink name strings, including zeros>
NumberOfInks 14E SHORT 1
DotRange 150 BYTE|SHORT 2|2*NumberOfInks
TargetPrinter 151 ASCII any
ExtraSamples 152 BYTE <number of extra components per pixel>
SampleFormat 153 SHORT SamplesPerPixel
SMinSampleValue 154 Any SamplesPerPixel
SMaxSampleValue 155 Any SamplesPerPixel
TransferRange 156 SHORT 6
JPEGProc 200 SHORT 1
JPEGInterchangeFormat 201 LONG 1
JPEGInterchangeFormatLength 202 LONG 1
JPEGRestartInterval 203 SHORT 1
JPEGLosslessPredictos 205 SHORT SamplesPerPixel
JPEGPointTransforms 206 SHORT SamplesPerPixel
JPEGQTables 207 LONG SamplesPerPixel
JPEGDCTables 208 LONG SamplesPerPixel
JPEGACTables 209 LONG SamplesPerPixel
YCbCrCoefficients 211 RATIONAL 3
YCbCrSubSampling 212 SHORT 2
YCbCrPositioning 213 SHORT 1
ReferenceBlackWhite 214 LONG 2*SamplesPerPixel
Copyright 8298 ASCII Any

# non-standard tags:
CZ_LSMInfo 866C CZ_LSM

# EXIF tags, see http://www.awaresystems.be/imaging/tiff/tifftags/privateifd/exif.html
EXIF_IFDOffset 8769 SHORT 1
EXIF_ExposureTime 829a RATIONAL 1
EXIF_FNumber 829d RATIONAL 1
EXIF_ExposureProgram 8822 SHORT 1
EXIF_SpectralSensitivity 8824 ASCII
EXIF_ISOSpeedRatings 8827 SHORT 1
EXIF_OECF 8828 UNDEFINED
EXIF_ExifVersion 9000 UNDEFINED 4
EXIF_DateTimeOriginal 9003 ASCII
EXIF_DateTimeDigitized 9004 ASCII
EXIF_ComponentsConfiguration 9101 UNDEFINED 4
EXIF_CompressedBitsPerPixel 9102 RATIONAL 1
EXIF_ShutterSpeedValue 9201 SRATIONAL 1
EXIF_ApertureValue 9202 RATIONAL 1
EXIF_BrightnessValue 9203 SRATIONAL 1
EXIF_ExposureBiasValue 9204 SRATIONAL 1
EXIF_MaxApertureValue 9205 RATIONAL 1
EXIF_SubjectDistance 9206 RATIONAL 1
EXIF_MeteringMode 9207 SHORT 1
EXIF_LightSource 9208 SHORT 1
EXIF_Flash 9209 SHORT 1
EXIF_FocalLength 920a RATIONAL 1
EXIF_SubjectArea 9214 SHORT 2|3|4
EXIF_MakerNote 927c UNDEFINED
EXIF_UserComment 9286 UNDEFINED
EXIF_SubsecTime 9290 ASCII
EXIF_SubsecTimeOriginal 9291 ASCII
EXIF_SubsecTimeDigitized 9292 ASCII
EXIF_FlashpixVersion a000 UNDEFINED 4
EXIF_ColorSpace a001 SHORT 1
EXIF_PixelXDimension a002 SHORT!LONG 1
EXIF_PixelYDimension a003 SHORT!LONG 1
EXIF_RelatedSoundFile a004 ASCII 13
EXIF_FlashEnergy a20b RATIONAL 1
EXIF_SpatialFrequencyResponse a20c UNDEFINED
EXIF_FocalPlaneXResolution a20e RATIONAL 1
EXIF_FocalPlaneYResolution a20f RATIONAL 1
EXIF_FocalPlaneResolutionUnit a210 SHORT 1
EXIF_SubjectLocation a214 SHORT 2
EXIF_ExposureIndex a215 RATIONAL 1
EXIF_SensingMethod a217 SHORT 1
EXIF_FileSource a300 UNDEFINED 1
EXIF_SceneType a301 UNDEFINED 1
EXIF_CFAPattern a302 UNDEFINED
EXIF_CustomRendered a401 SHORT 1
EXIF_ExposureMode a402 SHORT 1
EXIF_WhiteBalance a403 SHORT 1
EXIF_DigitalZoomRatio a404 RATIONAL 1
EXIF_FocalLengthIn35mmFilm a405 SHORT 1
EXIF_SceneCaptureType a406 SHORT 1
EXIF_GainControl a407 SHORT 1
EXIF_Contrast a408 SHORT 1
EXIF_Saturation a409 SHORT 1
EXIF_Sharpness a40a SHORT 1
EXIF_DeviceSettingDescription a40b UNDEFINED
EXIF_SubjectDistanceRange a40c SHORT 1
EXIF_ImageUniqueID a420 ASCII 33

'''

default_tag_values = dict(BitsPerSample=8, SampleFormat=1,
                          RowsPerStrip=2**32-1,
                          SamplesPerPixel=1, ExtraSamples = None,
                          PlanarConfiguration=1,
                          Compression=1, Predictor = 1, 
                          NewSubfileType=0,
                          Orientation = 1,
                          MaxSampleValue = None, MinSampleValue = None,
                          DateTime = None,
                          Artist = None,
                          HostComputer = None,
                          Software = None,
                          ImageDescription = None,
                          DocumentName = None,
                          ResolutionUnit = 2, XResolution = 1, YResolution = 1,
                          FillOrder = 1,
                          XPosition = None, YPosition = None,
                          Make = None, Model = None, Copyright=None,)

rational = numpy.dtype([('numer', numpy.uint32), ('denom', numpy.uint32)])
srational = numpy.dtype([('numer', numpy.int32), ('denom', numpy.int32)])

type2name = {1:'BYTE', 2:'ASCII', 3:'SHORT', 4:'LONG', 5:'RATIONAL', # two longs, lsm uses it for float64
             6:'SBYTE', 7:'UNDEFINED', 8:'SSHORT', 9:'SLONG', 10:'SRATIONAL',
             11:'FLOAT', 12:'DOUBLE',
             }
name2type = dict((v,k) for k,v in list(type2name.items()))
name2type['SHORT|LONG'] = name2type['LONG']
name2type['LONG|SHORT'] = name2type['LONG']
type2bytes = {1:1, 2:1, 3:2, 4:4, 5:8, 6:1, 7:1, 8:2, 9:4, 10:8, 11:4, 12:8}
type2dtype = {1:numpy.uint8, 2:numpy.uint8, 3:numpy.uint16, 4:numpy.uint32, 5:rational,
              6:numpy.int8, 8:numpy.int16, 9:numpy.int32,10:srational,
              11:numpy.float32, 12:numpy.float64}

tag_value2name = {}
tag_name2value = {}
tag_value2type = {}
for line in tag_info.split('\n'):
    if not line or line.startswith('#'): continue
    if line[0]==' ':
        pass
    else:
        n,h,t = line.split()[:3]
        h = eval('0x'+h)
        tag_value2name[h]=n
        tag_value2type[h]=t
        tag_name2value[n]=h

sample_format_map = {1:'uint', 2:'int', 3:'float', None:'uint', 6:'complex'}

class NumpyDTypes:

    def get_dtype(self, sample_format, bits_per_sample):
        format = sample_format_map[sample_format]
        dtypename = '%s%s' % (format, bits_per_sample)
        return getattr (self, dtypename)

class LittleEndianNumpyDTypes(NumpyDTypes):
    uint8 = numpy.dtype('<u1')
    uint16 = numpy.dtype('<u2')
    uint32 = numpy.dtype('<u4')
    uint64 = numpy.dtype('<u8')
    int8 = numpy.dtype('<i1')
    int16 = numpy.dtype('<i2')
    int32 = numpy.dtype('<i4')
    int64 = numpy.dtype('<i8')
    float32 = numpy.dtype('<f4')
    float64 = numpy.dtype('<f8')
    complex64 = numpy.dtype('<c8')
    complex128 = numpy.dtype('<c16')

    @property
    def type2dt(self):
        return dict((k,numpy.dtype(v).newbyteorder('<')) for k,v in list(type2dtype.items()))



LittleEndianNumpyDTypes = LittleEndianNumpyDTypes()

class BigEndianNumpyDTypes(NumpyDTypes):
    uint8 = numpy.dtype('>u1')
    uint16 = numpy.dtype('>u2')
    uint32 = numpy.dtype('>u4')
    uint64 = numpy.dtype('>u8')
    int8 = numpy.dtype('>i1')
    int16 = numpy.dtype('>i2')
    int32 = numpy.dtype('>i4')
    int64 = numpy.dtype('>i8')
    float32 = numpy.dtype('>f4')
    float64 = numpy.dtype('>f8')
    complex64 = numpy.dtype('>c8')
    complex128 = numpy.dtype('>c16')

    @property
    def type2dt(self):
        return dict((k,numpy.dtype(v).newbyteorder('>')) for k,v in list(type2dtype.items()))

BigEndianNumpyDTypes = BigEndianNumpyDTypes()
